/*
 * SignatureTM.cpp
 */

#include "SignatureTM.h"
#include "Conf.h"
#include "crypto/Digest.h"
#include "crypto/OpenSSLHelpers.h"
#include "crypto/cert/X509CertStore.h"
#include "crypto/ocsp/OCSP.h"
#include "util/DateTime.h"
#include "log.h"

/**
 * TM profile signature media type.
 */
const std::string digidoc::SignatureTM::MEDIA_TYPE = "signature/bdoc-1.0/TM";

digidoc::SignatureTM::SignatureTM(unsigned int id, const BDoc &_bdoc)
: SignatureBES(id, _bdoc)
{
}

digidoc::SignatureTM::SignatureTM(const std::string& path, const BDoc &_bdoc) throw(SignatureException)
: SignatureBES(path, _bdoc)
{
}

digidoc::SignatureTM::~SignatureTM()
{
}

/**
 * @return returns signature mimetype.
 */
std::string digidoc::SignatureTM::getMediaType() const
{
    return MEDIA_TYPE;
}

/**
 * @return nonce value
 */
std::vector<unsigned char> digidoc::SignatureTM::getNonce() const
{
    return digidoc::OCSP().getNonce(getOCSPResponseValue());
}

/**
 * @return returns OCSP certificate
 */
digidoc::X509Cert digidoc::SignatureTM::getOCSPCertificate() const
{
    const xades::UnsignedSignaturePropertiesType::CertificateValuesType::EncapsulatedX509CertificateSequence &certs =
        unsignedSignatureProperties()->certificateValues()[0].encapsulatedX509Certificate();
    xades::UnsignedSignaturePropertiesType::CertificateValuesType::EncapsulatedX509CertificateSequence::const_iterator i = certs.begin();

    try
    {
        for( ; i != certs.end(); ++i )
        {
            if( i->id() && i->id().get().find( "RESPONDER_CERT" ) != std::string::npos )
                return X509Cert( std::vector<unsigned char>( i->data(), i->data() + i->size() ) );
        }
        if( certs.begin() != certs.end() )
            return X509Cert( std::vector<unsigned char>(
                certs.begin()->data(), certs.begin()->data() + certs.begin()->size() ) );
    }
    catch( const Exception & ) {}
    return X509Cert();
}

/**
 * @return returns OCSP timestamp
 */
std::string digidoc::SignatureTM::getProducedAt() const
{
    std::vector<unsigned char> respBuf = getOCSPResponseValue();
    if(respBuf.empty())
        return "";
    tm datetime = OCSP().getProducedAt(respBuf);
    xml_schema::DateTime xmldatetime = digidoc::util::date::makeDateTime(datetime);
    return digidoc::util::date::xsd2string(xmldatetime);
}

/**
 * @return returns OCSP responder ID
 */
std::string digidoc::SignatureTM::getResponderID() const
{
    return unsignedSignatureProperties()->completeRevocationRefs()[0].oCSPRefs()
        ->oCSPRef()[0].oCSPIdentifier().responderID().byName().get();
}

/**
 * Do TM offline validations.
 * <ul>
 *   <li>Validate BES offline</li>
 *   <li>Check OCSP response (RevocationValues) was signed by trusted OCSP server</li>
 *   <li>Check that nonce field in OCSP response is same as CompleteRevocationRefs-&gt;DigestValue</li>
 *   <li>Recalculate hash of signature and compare with nonce</li>
 * </ul>
 * @throws SignatureException if signature is not valid
 */
void digidoc::SignatureTM::validateOffline() const throw(SignatureException)
{
    SignatureBES::validateOffline();

//       1. Check OCSP response (RevocationValues) was signed by OCSP server
//       2. OCSP server certificate is trusted?
//       3. Check that nonce field in OCSP response is same as CompleteRevocationRefs->DigestValue
//       4. Recalculate hash of signature and compare with nonce

#if 0
    Conf* conf = Conf::getInstance();
    Conf::OCSPConf ocspConf = conf->getOCSP(getSigningCertificate().getIssuerName());
    if(ocspConf.issuer.empty())
    {
        SignatureException e(__FILE__, __LINE__, "Failed to find ocsp responder.");
        e.setCode( Exception::OCSPResponderMissing );
        throw e;
    }
    OCSP ocsp(ocspConf.url);
    STACK_OF(X509)* ocspCerts = 0;
    try
    {
        ocspCerts = X509CertStore::getInstance()->findCerts( "CN", ocspConf.cert );
    }
    catch( const Exception &e )
    {
        SignatureException exception(__FILE__, __LINE__, "OCSP certificate loading failed", e);
        exception.setCode( Exception::OCSPCertMissing );
        throw exception;
	}
    X509Stack_scope x509StackScope(&ocspCerts);
    ocsp.setOCSPCerts(ocspCerts);
#else
    OCSP ocsp;
    ocsp.setCertStore(digidoc::X509CertStore::getInstance()->getCertStore());
    ocsp.setOCSPCerts(digidoc::X509CertStore::getInstance()->getCerts());
#endif
    std::vector<unsigned char> respBuf = getOCSPResponseValue();
    if(respBuf.empty())
        THROW_SIGNATUREEXCEPTION("OCSP response is empty");
    try
    {
        ocsp.verifyResponse(respBuf);
    }
    catch( const Exception &e )
    {
        THROW_SIGNATUREEXCEPTION_CAUSE( e, "OCSP response verfiy failed" );
    }
    DEBUG("OCSP response was signed by trusted OCSP responder");

    xades::UnsignedPropertiesType::UnsignedSignaturePropertiesOptional &usp = unsignedSignatureProperties();
    if(!usp.present())
        THROW_SIGNATUREEXCEPTION( "Failed to verify OCSP response producedAt" );
    xades::UnsignedSignaturePropertiesType::CompleteRevocationRefsSequence &crrs = usp->completeRevocationRefs();
    if(crrs.empty())
        THROW_SIGNATUREEXCEPTION( "Failed to verify OCSP response producedAt" );
    xades::CompleteRevocationRefsType::OCSPRefsOptional &orr = crrs[0].oCSPRefs();
    if(!orr.present())
        THROW_SIGNATUREEXCEPTION( "Failed to verify OCSP response producedAt" );
    xades::OCSPRefsType::OCSPRefSequence &ors = orr->oCSPRef();
    if(ors.empty())
        THROW_SIGNATUREEXCEPTION( "Failed to verify OCSP response producedAt" );
    xades::OCSPRefType::OCSPIdentifierType &oit = ors[0].oCSPIdentifier();
    if(oit.producedAt() != digidoc::util::date::makeDateTime(OCSP().getProducedAt(respBuf)))
        THROW_SIGNATUREEXCEPTION( "ProducedAt does not match with OCSP response" );

    std::vector<unsigned char> revocationOCSPRefValue(0);
    std::string method;
    getRevocationOCSPRef(revocationOCSPRefValue, method);

    std::auto_ptr<Digest> calc(new Digest(method));
    calc->update(getSignatureValue());
    std::vector<unsigned char> nonce = calc->getDigest();

    std::vector<unsigned char> respNonce = ocsp.getNonce(respBuf);
    if(nonce != respNonce)
    {
        DEBUGMEM("Calculated signature HASH", &nonce[0], nonce.size());
        DEBUGMEM("Response nonce", &respNonce[0], respNonce.size());
        THROW_SIGNATUREEXCEPTION("Calculated signature hash doesn't match to OCSP responder nonce field");
    }

    std::auto_ptr<Digest> ocspResponseCalc(new Digest(method));
    DEBUG("Calculating digest on %d bytes", respBuf.size());
    ocspResponseCalc->update(respBuf);
    std::vector<unsigned char> ocspResponseHash = ocspResponseCalc->getDigest();

    if(ocspResponseHash != revocationOCSPRefValue)
    {
        DEBUGMEM("Document   ocspResponse HASH:", &revocationOCSPRefValue[0], revocationOCSPRefValue.size());
        DEBUGMEM("Calculated ocspResponse HASH:", &ocspResponseHash[0], ocspResponseHash.size());
        THROW_SIGNATUREEXCEPTION("OCSPRef value doesn't match with hash of OCSP response");
    }
    else
    {
        DEBUG("TM signature valid");
    }
}

/**
 *
 * @param signer
 * @throws SignatureException
 */
void digidoc::SignatureTM::sign(Signer* signer) throw(SignatureException, SignException)
{
    DEBUG("SignatureTM::sign()");
    // Sign with BES profile.
    SignatureBES::sign(signer);
    DEBUG("BES signature successful.");
    notarize();
}

/**
 *
 * @throws SignatureException
 */
void digidoc::SignatureTM::notarize() throw(SignatureException, SignException)
{
    // Calculate NONCE value.
    std::auto_ptr<Digest> calc(new Digest());
    calc->update(getSignatureValue());
    std::vector<unsigned char> nonce = calc->getDigest();
    DEBUGMEM("Calculated signature HASH (nonce):", &nonce[0], nonce.size());

    // Get issuer certificate from certificate store.
    X509Cert cert = getSigningCertificate();
    X509* issuer = X509CertStore::getInstance()->getCert(*(cert.getIssuerNameAsn1())); X509_scope issuerScope(&issuer);
    if(issuer == NULL)
    {
        THROW_SIGNATUREEXCEPTION("Could not find certificate '%s' issuer '%s' in certificate store.",
            cert.getSubject().c_str(), cert.getIssuerName().c_str());
    }

    DEBUG("Signing with X.509 cert {serial=%s, subject=%s, issuer=%s})",
        cert.getSerial().c_str(), cert.getSubject().c_str(), cert.getIssuerName().c_str());

    // Initialize OCSP.
    DEBUG("Making OCSP request.");
    Conf* conf = Conf::getInstance();
    Conf::OCSPConf ocspConf = conf->getOCSP(cert.getIssuerName());
    if(ocspConf.issuer.empty())
    {
        SignatureException e(__FILE__, __LINE__, "Failed to find ocsp responder.");
        e.setCode( Exception::OCSPResponderMissing );
        throw e;
    }

    STACK_OF(X509)* ocspCerts = 0;
    try
    {
        ocspCerts = X509CertStore::getInstance()->findCerts( "CN", ocspConf.cert );
    }
    catch(const IOException& e)
    {
        THROW_SIGNATUREEXCEPTION_CAUSE(e, "Failed to load OCSP certificate");
    }
    X509Stack_scope ocspCertsScope(&ocspCerts);

    OCSP::CertStatus status = OCSP::UNKNOWN;
    std::vector<unsigned char> ocspResponse;
    struct tm producedAt;
    try
    {
        OCSP ocsp;
        ocsp.setOCSPCerts(ocspCerts);
        ocsp.setUrl(ocspConf.url);
        ocsp.setMaxAge(2*60); // FIXME: remove or move to conf
        ocsp.setSkew(15*60); // FIXME: remove or move to conf
        X509 *c = cert.getX509();
        status = ocsp.checkCert(c, issuer, nonce, ocspResponse, producedAt);
        X509_free(c);
    }
    catch(const IOException& e)
    {
        THROW_SIGNATUREEXCEPTION_CAUSE(e, "Failed to get OCSP response");
    }
    catch(const OCSPException& e)
    {
        THROW_SIGNATUREEXCEPTION_CAUSE(e, "Failed to get OCSP response");
    }

    switch(status)
    {
    case digidoc::OCSP::GOOD: DEBUG("OCSP status: GOOD"); break;
    case digidoc::OCSP::REVOKED:
    {
        DEBUG("OCSP status: REVOKED");
        SignatureException e( __FILE__, __LINE__, "Certificate status: revoked" );
        e.setCode( Exception::CertificateRevoked );
        throw e;
        break;
    }
    case digidoc::OCSP::UNKNOWN:
    {
        DEBUG("OCSP status: UNKNOWN");
        SignatureException e( __FILE__, __LINE__, "Certificate status: unknown" );
        e.setCode( Exception::CertificateUnknown );
        throw e;
        break;
    }
    }
    DEBUG("OCSP response size %d", ocspResponse.size());

    // FIXME: get from ocsp instead
    // FIXME: This file can contain multiple certs. X509Cert class supports only one cert per file
    // loadX509Stack loads multiple certs from one file
    // X509* ocspCert = X509Cert::loadX509(Conf::getInstance()->getOCSPCertPath()); X509_scope ocspCertScope(&ocspCert);
    if(sk_X509_num(ocspCerts) > 1)
    {
        ERR("More than one OCSP cert in file.");
    }
    X509Cert ocspCert_(sk_X509_value(ocspCerts, 0));

    std::auto_ptr<Digest> ocspResponseCalc(new Digest());
    ocspResponseCalc->update(ocspResponse);
    std::vector<unsigned char> ocspResponseHash = ocspResponseCalc->getDigest();
    DEBUGMEM("Calculated ocspResponse HASH:", &ocspResponseHash[0], ocspResponseHash.size());

    // Set TM profile signature parameters.
    createTMProperties();
    setOCSPCertificate(ocspCert_);
    setCACertificate(X509Cert(issuer));
    setOCSPResponseValue(ocspResponse);
    setCompleteRevocationRefs(ocspCert_.getIssuerName(), calc->getUri(), ocspResponseHash, producedAt);
}

/**
 * Add elements required by TM profile
 */
void digidoc::SignatureTM::createTMProperties()
{
    xades::CertificateValuesType certValue;

    xades::UnsignedSignaturePropertiesType usSignatureProp;
    usSignatureProp.certificateValues().push_back(certValue);

    xades::UnsignedPropertiesType usProp;
    usProp.unsignedSignatureProperties(usSignatureProp);

    signature->object()[0].qualifyingProperties()[0].unsignedProperties(usProp);
}


/**
 * Create CompleteRevocationRefs element that describes communication with OSCP responder.
 * @param responderName OCSP responder name as represented in responder public certification. Format as RFC2253
 * @param digestMethodUri digest method URI that was used for calculating ocspResponseHash
 * @param ocspResponseHash Digest of DER encode OCSP response
 * @param producedAt ProduceAt field of OCSP response
 */
void digidoc::SignatureTM::setCompleteRevocationRefs(const std::string& responderName, const std::string& digestMethodUri,
        const std::vector<unsigned char>& ocspResponseHash, const struct tm& producedAt )
{
    dsig::DigestMethodType method(xml_schema::Uri(digestMethodUri.c_str()));
    dsig::DigestValueType value(xml_schema::Base64Binary(&ocspResponseHash[0], ocspResponseHash.size()));

    xades::DigestAlgAndValueType digest(method, value);

    xades::ResponderIDType responderId;
    responderId.byName(responderName);

    xml_schema::DateTime dateTime(util::date::makeDateTime(producedAt));
    xades::OCSPIdentifierType ocspId(responderId, dateTime);
    std::string id = getId();
    id.replace(0, 1, "N");
    ocspId.uRI(xml_schema::Uri("#" + id));

    xades::OCSPRefType ocspRef(ocspId);
    ocspRef.digestAlgAndValue(digest);

    xades::OCSPRefsType ocspRefs;
    ocspRefs.oCSPRef().push_back(ocspRef);

    xades::CompleteRevocationRefsType completeRevocationRefs;
    completeRevocationRefs.oCSPRefs(ocspRefs);
    completeRevocationRefs.id(xml_schema::Id(getId() + "-REVOCREFS"));

    unsignedSignatureProperties()->completeRevocationRefs().push_back(completeRevocationRefs);
}

/**
 * Set OCSP responder cert.
 * @param x509 certificate that was used identify OCSP responder
 */
void digidoc::SignatureTM::setOCSPCertificate(const X509Cert& x509)
{
    //XXX: copied from digidoc::Signature::setSigningCertificate

    DEBUG("digidoc::SignatureTM::setOCSPCertificate()");

    std::vector<unsigned char> derEncodedX509 = x509.encodeDER();

    // Calculate SHA digest of the certificate.
    std::auto_ptr<Digest> calc(new Digest());
    calc->update(derEncodedX509);
    dsig::DigestMethodType digestMethod(xml_schema::Uri(calc->getUri()));
    dsig::DigestValueType digestValue(xml_schema::Base64Binary(&calc->getDigest()[0], calc->getSize()));
    xades::DigestAlgAndValueType certDigest(digestMethod, digestValue);

    // Add certificate issuer info.
    dsig::X509IssuerSerialType issuerSerial(x509.getIssuerName(), x509.getSerial());

    digidoc::xades::CertIDType cert(certDigest, issuerSerial);
    xades::CertIDListType certList;
    certList.cert().push_back(cert);

    xades::CompleteCertificateRefsType certificateRefs(certList);

    //certificateRefs.certRefs(certList);

    unsignedSignatureProperties()->completeCertificateRefs().push_back(certificateRefs);

    //CertificateValues
    addCertificateValue(getId() + "-RESPONDER_CERT", x509);
}

void digidoc::SignatureTM::setCACertificate(const X509Cert& x509)
{
    addCertificateValue(getId() + "-CA-CERT", x509);
}

/**
 * Add certificate under CertificateValues element
 * @param certId id attribute of EncapsulatedX509Certificate
 * @param x509 value of EncapsulatedX509Certificate
 */
void digidoc::SignatureTM::addCertificateValue(const std::string& certId, const X509Cert& x509)
{
    DEBUG("digidoc::SignatureTM::setCertificateValue(%s, X509Cert{%s,%s})",
        certId.c_str(), x509.getSerial().c_str(), x509.getSubject().c_str());
    //CertificateValues
    std::vector<unsigned char> certBytes = x509.encodeDER();
    xades::CertificateValuesType::EncapsulatedX509CertificateType certData(
        xml_schema::Base64Binary(&certBytes[0], certBytes.size()));
    certData.id(xml_schema::Id(certId.c_str()));
    unsignedSignatureProperties()->certificateValues()[0].encapsulatedX509Certificate().push_back(certData);
}

/**
 * Fill  RevocationValues element
 * @param data OCSP response in DER encoding
 */
void digidoc::SignatureTM::setOCSPResponseValue(const std::vector<unsigned char>& data)
{
    std::string id = getId();
    id.replace(0, 1, "N");
    xades::OCSPValuesType::EncapsulatedOCSPValueType ocspValueData(
        xml_schema::Base64Binary(&data[0], data.size()));
    ocspValueData.id(id);

    xades::OCSPValuesType ocspValue;
    ocspValue.encapsulatedOCSPValue().push_back(ocspValueData);

    xades::RevocationValuesType revocationValues;
    revocationValues.oCSPValues(ocspValue);

    unsignedSignatureProperties()->revocationValues().push_back(revocationValues);
}
/**
 * Get value of UnsignedProperties\UnsignedSignatureProperties\RevocationValues\OCSPValues\EncapsulatedOCSPValue
 * which contains whole OCSP response
 * @param data will contain DER encoded OCSP response bytes
 */
std::vector<unsigned char> digidoc::SignatureTM::getOCSPResponseValue() const
{
    try
    {
        //FIXME: check that all elements exist
        xades::RevocationValuesType &t = unsignedSignatureProperties()->revocationValues()[0];
        xades::OCSPValuesType &tt = t.oCSPValues().get();
        xades::OCSPValuesType::EncapsulatedOCSPValueType &resp = tt.encapsulatedOCSPValue()[0];
        return std::vector<unsigned char>(resp.data(), resp.data()+resp.size());
    }
    catch(const Exception &)
    {}
    return std::vector<unsigned char>();
}

/**
 * Get value of UnsignedProperties\UnsignedSignatureProperties\CompleteRevocationRefs\OCSPRefs\OCSPRef\DigestAlgAndValue
 * which contains hash of OCSP server response.
 * @param data Binary value of UnsignedProperties\UnsignedSignatureProperties\CompleteRevocationRefs\OCSPRefs\OCSPRef\DigestAlgAndValue\DigestValue. Hash of DER encoded OCSP response
 * @param digestMethodUri Digest method used for calculating hash of OCSP response
 * @throws SignatureException on error in references
 */
void digidoc::SignatureTM::getRevocationOCSPRef(std::vector<unsigned char>& data, std::string& digestMethodUri)
    const throw(SignatureException)
{
    xades::UnsignedSignaturePropertiesType::CompleteRevocationRefsSequence &crrSeq =
        unsignedSignatureProperties()->completeRevocationRefs();

    if(!crrSeq.empty())
    {
        xades::CompleteRevocationRefsType::OCSPRefsOptional &ocspRefsOpt = crrSeq[0].oCSPRefs();
        if(ocspRefsOpt.present())
        {
            xades::OCSPRefsType::OCSPRefSequence &ocspRefSeq = ocspRefsOpt->oCSPRef();
            if(!ocspRefSeq.empty())
            {
                xades::OCSPRefType::DigestAlgAndValueOptional &digestOpt = ocspRefSeq[0].digestAlgAndValue();
                if(digestOpt.present())
                {
                    dsig::DigestValueType &digestValue = digestOpt->digestValue();
                    data.resize(digestValue.size());
                    std::copy(digestValue.data(), digestValue.data()+digestValue.size(), data.begin());

                    xml_schema::Uri &uri = digestOpt->digestMethod().algorithm();
                    digestMethodUri = uri;

                    return;
                }
            }
        }
    }


    //XXX: improve exception message
    THROW_SIGNATUREEXCEPTION("Missing UnsignedProperties/UnsignedSignatureProperties/CompleteRevocationRefs/OCSPRefs/OCSPRef/DigestAlgAndValue element or its parent");
}

digidoc::xades::UnsignedPropertiesType::UnsignedSignaturePropertiesOptional&
digidoc::SignatureTM::unsignedSignatureProperties() const throw(SignatureException)
{
    // Object
    dsig::SignatureType::ObjectSequence &os = signature->object();
    if ( os.empty() )
        THROW_SIGNATUREEXCEPTION("Signature block 'Object' is missing.");
    else if ( os.size() != 1 )
        THROW_SIGNATUREEXCEPTION("Signature block contains more than one 'Object' block.");

    // QualifyingProperties
    dsig::ObjectType::QualifyingPropertiesSequence &qpSeq = os[0].qualifyingProperties();
    if ( qpSeq.empty() )
        THROW_SIGNATUREEXCEPTION("Signature block 'QualifyingProperties' is missing.");
    else if ( qpSeq.size() != 1 )
        THROW_SIGNATUREEXCEPTION("Signature block 'Object' contains more than one 'QualifyingProperties' block.");

    // UnsignedProperties
    xades::QualifyingPropertiesType::UnsignedPropertiesOptional &unsignedPropsOptional = qpSeq[0].unsignedProperties();
    if ( !unsignedPropsOptional.present() )
        THROW_SIGNATUREEXCEPTION("QualifyingProperties block 'UnsignedProperties' is missing.");

    // UnsignedSignatureProperties
    digidoc::xades::UnsignedPropertiesType::UnsignedSignaturePropertiesOptional &unsignedSigProps =
    unsignedPropsOptional->unsignedSignatureProperties();
    if ( !unsignedSigProps.present() )
        THROW_SIGNATUREEXCEPTION("QualifyingProperties block 'UnsignedSignatureProperties' is missing.");

    return unsignedSigProps;
}
