/*
 * XmlConf.cpp
 */

#include "XmlConf.h"

#include "log.h"
#include "util/File.h"
#include "util/File_p.h"
#include "xml/conf.hxx"

#include <xsd/cxx/xml/dom/parsing-source.hxx> //for parsing xml data into DOMDocument

#include <iostream>
#include <fstream>
#ifdef _WIN32
#include <windows.h>
#endif

namespace digidoc
{

class XmlErrorHandler: public ::xsd::cxx::xml::error_handler< char >
{
public:
    virtual bool
    handle (const std::basic_string< char >&,
            unsigned long line,
            unsigned long column,
            severity,
            const std::basic_string< char >& message)
    {
        DEBUG("Failed to parse (line %u, column %u): %s",
            line, column, message.c_str());
        return false;
    }
};

template <class A>
class XmlConfParam
{
public:
    XmlConfParam(): locked(false) {}
    XmlConfParam(A val): value(val), locked(false) {}

    void setValue(const A &val, bool lock, bool global)
    {
        if(global) locked = lock;
        if(global || !locked) value = val;
    }

    XmlConfParam & operator=( const A &other )
    {
        value = other; return *this;
    }

    A value;
    bool locked;
};

class XmlConfPrivate
{
public:
    XmlConfPrivate()
        : logLevel(Log::InfoType)
        , pkcs12Disable(false)
    {
        setDefaultConfPath();
        setUserConfPath();
    }

    void init(const std::string &path, bool global) throw(IOException);
    std::auto_ptr<Configuration> read(const std::string &path) throw(IOException);

    void setUserConf(const std::string &paramName, const std::string &value) throw(IOException);
    void serializeUserConf(const Configuration &pConf) throw(IOException);

    static std::string getConfSchemaLocation() throw(IOException);
    static void setDefaultConfPath();
    static void setUserConfPath();

    XmlConfParam<int> logLevel;
    XmlConfParam<std::string> logFile;
    XmlConfParam<std::string> digestUri;
    XmlConfParam<std::string> signatureUri;
    XmlConfParam<std::string> pkcs11DriverPath;
    XmlConfParam<std::string> certStorePath;
    XmlConfParam<std::string> manifestXsdPath;
    XmlConfParam<std::string> xadesXsdPath;
    XmlConfParam<std::string> dsigXsdPath;
    XmlConfParam<std::string> proxyHost;
    XmlConfParam<std::string> proxyPort;
    XmlConfParam<std::string> proxyUser;
    XmlConfParam<std::string> proxyPass;
    XmlConfParam<std::string> pkcs12Cert;
    XmlConfParam<std::string> pkcs12Pass;
    XmlConfParam<bool> pkcs12Disable;
    std::vector<XmlConf::OCSPConf> ocsp;

    static std::string DEFAULT_CONF_LOC;
    static std::string USER_CONF_LOC;

    static const std::string LOG_LEVEL;
    static const std::string LOG_FILE;
    static const std::string DIGEST_URI;
    static const std::string SIGNATURE_URI;
    static const std::string PKCS11_DRIVER_PATH;
    static const std::string CERT_STORE_PATH;
    static const std::string MANIFEST_XSD_PATH;
    static const std::string XADES_XSD_PATH;
    static const std::string DSIG_XSD_PATH;
    static const std::string PROXY_HOST;
    static const std::string PROXY_PORT;
    static const std::string PROXY_USER;
    static const std::string PROXY_PASS;
    static const std::string PKCS12_CERT;
    static const std::string PKCS12_PASS;
    static const std::string PKCS12_DISABLE;
};
}

using namespace digidoc;

#ifdef _WIN32
#ifndef DIGIDOCPP_PATH_REGISTRY_KEY
#define DIGIDOCPP_PATH_REGISTRY_KEY "SOFTWARE\\Estonian ID Card\\digidocpp"
#endif
#endif

/**
 * Path to default configuration files
 */
std::string XmlConfPrivate::DEFAULT_CONF_LOC = "";
std::string XmlConfPrivate::USER_CONF_LOC = "";

const std::string XmlConfPrivate::LOG_LEVEL          = "log.level";
const std::string XmlConfPrivate::LOG_FILE           = "log.file";
const std::string XmlConfPrivate::DIGEST_URI         = "digest.uri";
const std::string XmlConfPrivate::SIGNATURE_URI      = "signature.uri";
const std::string XmlConfPrivate::PKCS11_DRIVER_PATH = "pkcs11.driver.path";
const std::string XmlConfPrivate::CERT_STORE_PATH    = "cert.store.path";
const std::string XmlConfPrivate::MANIFEST_XSD_PATH  = "manifest.xsd.path";
const std::string XmlConfPrivate::XADES_XSD_PATH     = "xades.xsd.path";
const std::string XmlConfPrivate::DSIG_XSD_PATH      = "dsig.xsd.path";
const std::string XmlConfPrivate::PROXY_HOST         = "proxy.host";
const std::string XmlConfPrivate::PROXY_PORT         = "proxy.port";
const std::string XmlConfPrivate::PROXY_USER         = "proxy.user";
const std::string XmlConfPrivate::PROXY_PASS         = "proxy.pass";
const std::string XmlConfPrivate::PKCS12_CERT        = "pkcs12.cert";
const std::string XmlConfPrivate::PKCS12_PASS        = "pkcs12.pass";
const std::string XmlConfPrivate::PKCS12_DISABLE     = "pkcs12.disable";

/**
 * Parses xml configuration given path
 * @param path to parse xml config
 * @return returns parsed xml configuration
 */
std::auto_ptr<Configuration> XmlConfPrivate::read(const std::string &path) throw(IOException)
{
    try
    {
        return configuration(path, xml_schema::Flags::dont_initialize);
    }
    catch(const xml_schema::Exception& e)
    {
        std::ostringstream oss;
        oss << e;
        THROW_IOEXCEPTION("Failed to parse configuration: %s - %s", path.c_str(), oss.str().c_str());
    }
    return std::auto_ptr<Configuration>(new Configuration);
}

/**
 * Use digidoc::XmlConf as configuration instance.
 */
void XmlConf::initialize()
{
    if(!Conf::isInitialized())
        Conf::init(new XmlConf());
}


/**
 * Tries to initialize XmlConf by using file defined in DIGIDOCPP_OVERRIDE_CONF environment variable.
 * If this is undefined, tries to load configuration from defined Default and user configuration file
 */
XmlConf::XmlConf() throw(IOException)
: d(new XmlConfPrivate)
{
    try
    {
        d->init(XmlConfPrivate::DEFAULT_CONF_LOC, true);
    }
    catch(const IOException &)
    {}

    try
    {
        d->init(XmlConfPrivate::USER_CONF_LOC, false);
    }
    catch(const IOException &)
    {
        // remove invalid user home config (wrong schema location, etc.)
        try
        {
            util::File::removeFile(XmlConfPrivate::USER_CONF_LOC);
        }
        catch(const IOException &)
        {
            THROW_IOEXCEPTION("Failed to remove invalid user home configuration '%s' file",XmlConfPrivate::USER_CONF_LOC.c_str());
        }
    }
}

/**
 * Initialize xml conf from path
 * @param path to use for initializing conf
 * @return
 */
XmlConf::XmlConf(const std::string& path) throw(IOException)
: d(new XmlConfPrivate)
{
    d->init(path, true);
}

XmlConf::~XmlConf()
{
    delete d;
}

/**
 * Gets default configuration file directory.
 * @return default configuration file directory full path.
 */
std::string XmlConf::getDefaultConfDir()
{
    // the file path in conf is relative to the conf file's location
    if( !XmlConfPrivate::DEFAULT_CONF_LOC.empty() )
        return util::File::directory( XmlConfPrivate::DEFAULT_CONF_LOC );
    return util::File::cwd();
}

/**
 * Load and parse xml from path. Initialize XmlConf member variables from xml.
 * @param path to use for initializing conf
 */
void XmlConfPrivate::init(const std::string& path, bool global) throw(IOException)
{
    DEBUG("digidoc::XmlConf::init(%s)", path.c_str());
    std::auto_ptr<Configuration> conf = read(path);
    try
    {
        Configuration::ParamSequence paramSeq = conf->param();
        for( Configuration::ParamSequence::const_iterator it = paramSeq.begin(); it != paramSeq.end(); it++)
        {
            if(LOG_LEVEL.compare(it->name()) == 0)
                logLevel.setValue(atoi(std::string(*it).c_str()), it->lock(), global);
            else if(LOG_FILE.compare(it->name()) == 0)
                logFile.setValue(*it, it->lock(), global);
            else if(DIGEST_URI.compare(it->name()) == 0)
                digestUri.setValue(*it, it->lock(), global);
            else if(SIGNATURE_URI.compare(it->name()) == 0)
                signatureUri.setValue(*it, it->lock(), global);
            else if(MANIFEST_XSD_PATH.compare(it->name()) == 0)
                manifestXsdPath.setValue(*it, it->lock(), global);
            else if(XADES_XSD_PATH.compare(it->name()) == 0)
                xadesXsdPath.setValue(*it, it->lock(), global);
            else if(DSIG_XSD_PATH.compare(it->name()) == 0)
                dsigXsdPath.setValue(*it, it->lock(), global);
            else if(PKCS11_DRIVER_PATH.compare(it->name()) == 0)
                pkcs11DriverPath.setValue(*it, it->lock(), global);
            else if(CERT_STORE_PATH.compare(it->name()) == 0)
                certStorePath.setValue(*it, it->lock(), global);
            else if(PROXY_HOST.compare(it->name()) == 0)
                proxyHost.setValue(*it, it->lock(), global);
            else if(PROXY_PORT.compare(it->name()) == 0)
                proxyPort.setValue(*it, it->lock(), global);
            else if(PROXY_USER.compare(it->name()) == 0)
                proxyUser.setValue(*it, it->lock(), global);
            else if(PROXY_PASS.compare(it->name()) == 0)
                proxyPass.setValue(*it, it->lock(), global);
            else if(PKCS12_CERT.compare(it->name()) == 0)
                pkcs12Cert.setValue(*it, it->lock(), global);
            else if(PKCS12_PASS.compare(it->name()) == 0)
                pkcs12Pass.setValue(*it, it->lock(), global);
            else if(PKCS12_DISABLE.compare(it->name()) == 0)
                pkcs12Disable.setValue((*it).compare("true") == 0, it->lock(), global);
            else
                WARN("Unknown configuration parameter %s", it->name().c_str());
        }

        Configuration::OcspSequence ocspSeq = conf->ocsp();
        for( Configuration::OcspSequence::const_iterator it = ocspSeq.begin(); it != ocspSeq.end(); ++it)
        {
            Conf::OCSPConf o;
            o.issuer = it->issuer();
            o.url = it->url();
            o.cert = it->cert();
            ocsp.push_back(o);
        }
    }
    catch(const xml_schema::Exception& e)
    {
        std::ostringstream oss;
        oss << e;
        THROW_IOEXCEPTION("Failed to parse configuration: %s", oss.str().c_str());
    }
}

/**
 * Sets Default (global) configuration file path in DEFAULT_CONF_LOC variable.
 */
void XmlConfPrivate::setDefaultConfPath()
{
    DEFAULT_CONF_LOC.clear();
    std::string overrideConf = util::File::env( "DIGIDOCPP_OVERRIDE_CONF" );
    if (!overrideConf.empty()) //if there is environment variable defined, use this conf instead of others
        DEFAULT_CONF_LOC = overrideConf;
    else
    {
#ifdef _WIN32
        HKEY hkey;
        if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT(DIGIDOCPP_PATH_REGISTRY_KEY), 0, KEY_QUERY_VALUE, &hkey)==ERROR_SUCCESS)
        {
            DWORD dwSize = MAX_PATH * sizeof(TCHAR);
            TCHAR tcConfPath[MAX_PATH];
            DWORD dwRet = RegQueryValueEx(hkey, TEXT("ConfigFile"), NULL, NULL, (LPBYTE)tcConfPath, &dwSize);
            RegCloseKey(hkey);
            if (dwRet == ERROR_SUCCESS)
                DEFAULT_CONF_LOC = util::File::decodeName(tcConfPath);
            else
                DEBUG("Failed to open registry key \"%s\" ConfigFile value ", DIGIDOCPP_PATH_REGISTRY_KEY);
        }
#else
        DEFAULT_CONF_LOC = DIGIDOCPP_CONFIG_DIR "/digidocpp.conf";
#endif
    }
}

/**
 * Sets User specific configuration file path in USER_CONF_LOC variable.
 */
void XmlConfPrivate::setUserConfPath()
{
#ifdef _WIN32
    USER_CONF_LOC = util::File::env("APPDATA");
    if (!USER_CONF_LOC.empty())
        USER_CONF_LOC += "\\digidocpp\\digidocpp.conf";
#else
    USER_CONF_LOC = util::File::env("HOME");
    if (!USER_CONF_LOC.empty())
        USER_CONF_LOC += "/.digidocpp/digidocpp.conf";
#endif
}

/**
 * Gets log level.
 * @return log level.
 */
int XmlConf::getLogLevel() const
{
    return d->logLevel.value;
}

/**
 * Gets log file location.
 * @return log path location.
 */
std::string XmlConf::getLogFile() const
{
    return d->logFile.value.empty() ? Conf::getLogFile() : d->logFile.value;
}

/**
 * Digest method used to calculate digest values in BDOC (e.g. 'http://www.w3.org/2000/09/xmldsig#sha1' for SHA1)
 *
 * For available method URIs see:
 * <li>
 *   <ul><b>W3C XML Encryption Syntax and Processing</b> (10 December 2005) http://www.w3.org/TR/xmlenc-core/</ul>
 *   <ul><b>RFC 4051</b> http://www.ietf.org/rfc/rfc4051.txt</ul>
 * </li>
 *
 * @return
 */
std::string XmlConf::getDigestUri() const
{
    return d->digestUri.value.empty() ? Conf::getDigestUri() : d->digestUri.value;
}

/**
 * Signature method used to calculate signature value in BDOC (e.g. 'http://www.w3.org/2000/09/xmldsig#rsa-sha1' for SHA1)
 *
 * For available method URIs see:
 * <li>
 *   <ul><b>W3C XML Encryption Syntax and Processing</b> (10 December 2005) http://www.w3.org/TR/xmlenc-core/</ul>
 *   <ul><b>RFC 4051</b> http://www.ietf.org/rfc/rfc4051.txt</ul>
 * </li>
 *
 * @return
 */
std::string XmlConf::getSignatureUri() const
{
    return d->signatureUri.value.empty() ? Conf::getSignatureUri() : d->signatureUri.value;
}


/**
 * Gets user specific configuration file directory.
 * @return returns user configuration file directory.
 */
std::string XmlConf::getUserConfDir()
{
    return util::File::directory(XmlConfPrivate::USER_CONF_LOC);
}

/**
 * Gets Manifest schema file location.
 * @return Manifest schema full path location.
 */
std::string XmlConf::getManifestXsdPath() const
{
    return util::File::fullPathUrl(getDefaultConfDir(),
        d->manifestXsdPath.value.empty() ? Conf::getManifestXsdPath() : d->manifestXsdPath.value);
}

/**
 * Gets Xades schema file location.
 * @return Xades schema full path location.
 */
std::string XmlConf::getXadesXsdPath() const
{
    return util::File::fullPathUrl(getDefaultConfDir(),
        d->xadesXsdPath.value.empty() ? Conf::getXadesXsdPath() : d->xadesXsdPath.value);
}

/**
 * Gets Dsig schema file location.
 * @return Dsig schema full path location.
 */
std::string XmlConf::getDsigXsdPath() const
{
    return util::File::fullPathUrl(getDefaultConfDir(),
        d->dsigXsdPath.value.empty() ? Conf::getDsigXsdPath() : d->dsigXsdPath.value);
}

/**
 * Gets PKCS11 driver file path.
 * @return PKCS11 driver file location.
 */
std::string XmlConf::getPKCS11DriverPath() const
{
    return d->pkcs11DriverPath.value.empty() ? Conf::getPKCS11DriverPath() : d->pkcs11DriverPath.value;
}

/**
 * Gets OCSP data by issuer.
 * @param issuer OCSP issuer.
 * @return returns OCSP data structure, containing issuer, url and certificate location.
 */
Conf::OCSPConf XmlConf::getOCSP(const std::string &issuer) const
{
    size_t pos = issuer.find( "CN=", 0 ) + 3;
    std::string _issuer = issuer.substr( pos, issuer.find( ",", pos ) - pos );
    for(std::vector<OCSPConf>::const_iterator i = d->ocsp.begin(); i != d->ocsp.end(); ++i)
    {
        if(i->issuer == _issuer)
            return *i;
    }
    return Conf::getOCSP(_issuer);
}

/**
 * Gets Certificate store location.
 * @return Certificate store full path location.
 */
std::string XmlConf::getCertStorePath() const
{
    std::string path = d->certStorePath.value.empty() ? Conf::getCertStorePath() : d->certStorePath.value;
    if( path.empty() ) return path;
    return util::File::isRelative( path ) ? util::File::path( getDefaultConfDir(), path ) : path;
}

/**
 * Gets proxy host address.
 * @return proxy host address.
 */
std::string XmlConf::getProxyHost() const
{
    return d->proxyHost.value.empty() ? Conf::getProxyHost() : d->proxyHost.value;
}

/**
 * Gets proxy port number.
 * @return proxy port.
 */
std::string XmlConf::getProxyPort() const
{
    return d->proxyPort.value.empty() ? Conf::getProxyPort() : d->proxyPort.value;
}

/**
 * Gets proxy user name.
 * @return proxy user name.
 */
std::string XmlConf::getProxyUser() const
{
    return d->proxyUser.value.empty() ? Conf::getProxyUser() : d->proxyUser.value;
}

/**
 * Gets proxy login password.
 * @return proxy password.
 */
std::string XmlConf::getProxyPass() const
{
    return d->proxyPass.value.empty() ? Conf::getProxyPass() : d->proxyPass.value;
}

/**
 * Gets PKCS12 certificate file location.
 * @return PKCS12 certificate full path location.
 */
std::string XmlConf::getPKCS12Cert() const
{
    return d->pkcs12Cert.value.empty() ? Conf::getPKCS12Cert() : d->pkcs12Cert.value;
}

/**
 * Gets PKCS12 password.
 * @return PKCS12 password.
 */
std::string XmlConf::getPKCS12Pass() const
{
    return d->pkcs12Pass.value.empty() ? Conf::getPKCS12Pass() : d->pkcs12Pass.value;
}

/**
 * Gets PKCS12 usage.
 * @return PKCS12 usage.
 */
bool XmlConf::getPKCS12Disable() const
{
    return d->pkcs12Disable.value;
}

/**
 * Sets a Proxy host address. Also adds or replaces proxy host data in the user configuration file.
 *
 * @param host proxy host address.
 * @throws IOException exception is thrown if saving a proxy host address into a user configuration file fails.
 */
void XmlConf::setProxyHost( const std::string &host ) throw(IOException)
{
    if( d->proxyHost.locked )
        return;
    d->proxyHost = host;
    d->setUserConf(XmlConfPrivate::PROXY_HOST, host);
}

/**
 * Sets a Proxy port number. Also adds or replaces proxy port data in the user configuration file.
 *
 * @param port proxy port number.
 * @throws IOException exception is thrown if saving a proxy port number into a user configuration file fails.
 */
void XmlConf::setProxyPort( const std::string &port ) throw(IOException)
{
    if( d->proxyPort.locked )
        return;
    d->proxyPort = port;
    d->setUserConf(XmlConfPrivate::PROXY_PORT, port);
}

/**
 * Sets a Proxy user name. Also adds or replaces proxy user name in the user configuration file.
 *
 * @param user proxy user name.
 * @throws IOException exception is thrown if saving a proxy user name into a user configuration file fails.
 */
void XmlConf::setProxyUser( const std::string &user ) throw(IOException)
{
    if( d->proxyUser.locked )
        return;
    d->proxyUser = user;
    d->setUserConf(XmlConfPrivate::PROXY_USER, user);
}

/**
 * Sets a Proxy password. Also adds or replaces proxy password in the user configuration file.
 *
 * @param pass proxy password.
 * @throws IOException exception is thrown if saving a proxy password into a user configuration file fails.
 */
void XmlConf::setProxyPass( const std::string &pass ) throw(IOException)
{
    if( d->proxyPass.locked )
        return;
    d->proxyPass = pass;
    d->setUserConf(XmlConfPrivate::PROXY_PASS, pass);
}

/**
 * Sets a PKCS#12 certficate path. Also adds or replaces PKCS#12 certificate path in the user configuration file.
 * By default the PKCS#12 certificate file should be located at default path, given by getUserConfDir() function.
 *
 * @param cert PKCS#12 certificate location path.
 * @throws IOException exception is thrown if saving a PKCS#12 certificate path into a user configuration file fails.
 */
void XmlConf::setPKCS12Cert( const std::string &cert ) throw(IOException)
{
    if( d->pkcs12Cert.locked )
        return;
    d->pkcs12Cert = cert;
    d->setUserConf(XmlConfPrivate::PKCS12_CERT, cert);
}

/**
 * Sets a PKCS#12 certificate password. Also adds or replaces PKCS#12 certificate password in the user configuration file.
 *
 * @param pass PKCS#12 certificate password.
 * @throws IOException exception is thrown if saving a PKCS#12 certificate password into a user configuration file fails.
 */
void XmlConf::setPKCS12Pass( const std::string &pass ) throw(IOException)
{
    if( d->pkcs12Pass.locked )
        return;
    d->pkcs12Pass = pass;
    d->setUserConf(XmlConfPrivate::PKCS12_PASS, pass);
}

/**
 * Sets a PKCS#12 certificate usage. Also adds or replaces PKCS#12 certificate usage in the user configuration file.
 *
 * @param pass PKCS#12 certificate usage.
 * @throws IOException exception is thrown if saving a PKCS#12 certificate usage into a user configuration file fails.
 */
void XmlConf::setPKCS12Disable( bool disable ) throw(IOException)
{
    if( d->pkcs12Disable.locked )
        return;
    d->pkcs12Disable = disable;
    d->setUserConf(XmlConfPrivate::PKCS12_DISABLE, disable ? "true" : "false");
}

/**
 * Adds or replaces OCSP parameters in the user configuration file.
 *
 * @param issuer an ocsp certificate issuer.
 * @param url an OCSP server url.
 * @param cert an OCSP certificate CN.
 * @throws IOException exception is thrown if saving an OCSP parameters into a user configuration file fails.
 */
void XmlConf::setOCSP(const std::string &issuer, const std::string &url, const std::string &cert) throw(IOException)
{
    std::auto_ptr<Configuration> conf(new Configuration);
    if(util::File::fileExists(XmlConfPrivate::USER_CONF_LOC))
        conf = d->read(XmlConfPrivate::USER_CONF_LOC);
    try
    {
        Configuration::OcspSequence ocspSeq = conf->ocsp();
        for(Configuration::OcspSequence::iterator it = ocspSeq.begin(); it != ocspSeq.end(); it++)
        {
            if (issuer.compare(it->issuer()) == 0)
            {
                ocspSeq.erase(it);
                break;
            }
        }
        if(url.size() || cert.size()) //if it's a new parameter
            ocspSeq.push_back(Ocsp(url, cert, issuer));
        conf->ocsp(ocspSeq); //replace all ocsp data with new modified ocsp sequence
    }
    catch(const xml_schema::Exception& e)
    {
        std::ostringstream oss;
        oss << e;
        THROW_IOEXCEPTION("(in set OCSP) Failed to parse configuration: %s", oss.str().c_str());
    }
    d->serializeUserConf(*conf);
}

/**
 * Sets any parameter in a user configuration file. Also creates a configuration file if it is missing.
 *
 * @param paramName name of parameter that needs to be changed or created.
 * @param value value for changing or adding to a given parameter. If value is an empty string, the walue for a given parameter will be erased.
 * @throws IOException exception is thrown if reading, writing or creating of a user configuration file fails.
 */
void XmlConfPrivate::setUserConf(const std::string &paramName, const std::string &value) throw(IOException)
{
    std::auto_ptr<Configuration> conf(new Configuration);
    if(util::File::fileExists(USER_CONF_LOC))
        conf = read(USER_CONF_LOC);
    try
    {
        Configuration::ParamSequence paramSeq = conf->param();
        for(Configuration::ParamSequence::iterator it = paramSeq.begin(); it != paramSeq.end(); it++)
        {
            if(paramName.compare(it->name()) == 0)
            {
                paramSeq.erase(it);
                break;
            }
        }
        if(value.size()) //if it's a new parameter
            paramSeq.push_back(Param(value, paramName));
        conf->param(paramSeq); //replace all param data with new modified param sequence
    }
    catch (const xml_schema::Exception& e)
    {
        std::ostringstream oss;
        oss << e;
        THROW_IOEXCEPTION("(in set %s) Failed to parse configuration: %s", paramName.c_str(), oss.str().c_str());
    }
    serializeUserConf(*conf);
}

/**
 * Writes configuration data to user specific configuration xml file.
 *
 * @param pConf configuration data in xsd object using conf.xsd schema.
 * @throws IOException exception is thrown if opening of user configuration file fails.
 */
void XmlConfPrivate::serializeUserConf(const Configuration &pConf) throw(IOException)
{
    if (!util::File::directoryExists(XmlConf::getUserConfDir()))
        util::File::createDirectory(XmlConf::getUserConfDir());
    f_string enc = util::File::encodeName(USER_CONF_LOC);
    std::ofstream ofs(enc.c_str());
    if (ofs.fail())
        THROW_IOEXCEPTION("Failed to open configuration: %s", USER_CONF_LOC.c_str());
    xml_schema::NamespaceInfomap map;
    map[""].name = "";
    map[""].schema = getConfSchemaLocation();
    configuration(ofs, pConf, map);
}

/**
 * Gets configuration file schema location with full path.
 * @throws IOException exception is thrown if failed to parse schema location from default conf file.
 * @return returns configuration file schema location.
 */
std::string XmlConfPrivate::getConfSchemaLocation() throw(IOException)
{
    try
    {
        //Parse default configuration file into a DOM document
        XmlErrorHandler errH;
        xml_schema::dom::auto_ptr< xercesc::DOMDocument > domDoc =
            ::xsd::cxx::xml::dom::parse< char > ( DEFAULT_CONF_LOC, errH,
                xml_schema::Properties(), xml_schema::Flags::dont_initialize );
        //Get access to root element
        if( domDoc.get() )
        {
            xercesc::DOMElement *root = domDoc->getDocumentElement();
            //Get schema location atributes.
            const XMLCh *xsl = root->getAttributeNS (
                xercesc::SchemaSymbols::fgURI_XSI,
                xercesc::SchemaSymbols::fgXSI_NONAMESPACESCHEMALOCACTION);
            return util::File::path( XmlConf::getDefaultConfDir(),
                ::xsd::cxx::xml::transcode<char> (xsl) );
        }
    }
    catch (...)
    {
        DEBUG("Failed to parse schema location in: %s", DEFAULT_CONF_LOC.c_str());
    }

    return util::File::path( XmlConf::getDefaultConfDir(), "schema/conf.xsd" );
}
