#include "DirectoryX509CertStore.h"
#include "X509CertStore_p.h"

#include "../../Conf.h"
#include "../../log.h"
#include "../../util/File.h"

#include <openssl/err.h>

/**
 * Gets certificate store path from configuration and loads all certificates
 * found in directory and adds these to the certificate store.
 *
 * @throws IOException exception is throws if the folder does not exist.
 */
digidoc::DirectoryX509CertStore::DirectoryX509CertStore() throw(IOException)
{
    loadCerts(Conf::getInstance()->getCertStorePath());
}

/**
 * Load all certificates found in directory and adds these to the cert store.
 *
 * @param path path to X.509 certificates in PEM format.
 * @throws IOException exception is throws if the folder does not exist.
 */
digidoc::DirectoryX509CertStore::DirectoryX509CertStore(const std::string &path) throw(IOException)
{
    loadCerts(path);
}

/**
 * Load all certificates found in directory and adds these to the cert store.
 *
 * @param path path to X.509 certificates in PEM format.
 * @throws IOException exception is throws if the folder does not exist.
 */
void digidoc::DirectoryX509CertStore::loadCerts(const std::string &path) throw(IOException)
{
    if(!util::File::directoryExists(path))
        THROW_IOEXCEPTION("Directory %s does not exists, can not load cert store.", path.c_str());

    std::vector<std::string> files = util::File::listFiles(path);
    for(std::vector<std::string>::const_iterator iter = files.begin(); iter != files.end(); iter++)
    {
        try
        {
            X509 *c = X509Cert::loadX509(*iter);
            sk_X509_push(d->stack, c);
            if(!X509_STORE_add_cert(d->store, c))
                WARN("Cant add cert %ld to X509_STORE, %s", ASN1_INTEGER_get(X509_get_serialNumber(c)), ERR_reason_error_string(ERR_get_error()));
        }
        catch(const IOException& e)
        {
            WARN(e.getMsg().c_str());
        }
    }
    INFO("Loaded %d certificates into certificate store.", sk_X509_num(d->stack));
}
