#include "X509CertStore.h"

#include "../../log.h"
#include "../../crypto/OpenSSLHelpers.h"

#include <sstream>
#include <string.h>

#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include <openssl/x509v3.h>

/**
 * Creates NULL X509 certificate.
 */
digidoc::X509Cert::X509Cert(): cert(NULL) {}

/**
 * Creates copy of the X509 certificate.
 *
 * @param cert X509 certificate structure to be wrapped.
 * @throws IOException throws exception if the X509 certificate structure copy fails.
 */
digidoc::X509Cert::X509Cert(X509* cert) throw(IOException)
 : cert(NULL)
{
    this->cert = copyX509(cert);
}

/**
 * Creates X509 certificate from the provider DER encoded bytes.
 *
 * @param bytes X509 certificate in DER encoding.
 * @throws IOException throws exception if X509 certificate parsing failed.
 */
digidoc::X509Cert::X509Cert(std::vector<unsigned char> bytes) throw(IOException)
 : cert(NULL)
{
    if(bytes.empty())
    {
        THROW_IOEXCEPTION("No bytes given to parse X509.");
    }

    // Parse certificate from DER formatted buffer.
    const unsigned char* pBytes = reinterpret_cast<const unsigned char*>(&bytes[0]);
    d2i_X509(&cert, &pBytes, (unsigned int)bytes.size());
    if(cert == NULL)
    {
        THROW_IOEXCEPTION("Failed to parse X509 certificate from bytes given: %s", ERR_reason_error_string(ERR_get_error()));
    }
}

/**
 * Copy constructor.
 *
 * @param copy instance of X509Cert class to be copied.
 * @throws IOException throws exception if the X509 cert structure copy fails.
 */
digidoc::X509Cert::X509Cert(const X509Cert& copy) throw(IOException)
 : cert(NULL)
{
    this->cert = copyX509(copy.cert);
}

/**
 * Clean up underlying X509* struct.
 */
digidoc::X509Cert::~X509Cert()
{
    X509_free(cert);
}

/**
 * Assignment operator.
 *
 * @param copy instance of X509Cert class to be assigned.
 * @return copies parameters from the copy instance.
 * @throws IOException throws exception if the X509 cert structure copy fails.
 */
digidoc::X509Cert& digidoc::X509Cert::operator=(const X509Cert& copy) throw(IOException)
{
    this->cert = copyX509(copy.cert);
    return *this;
}

/**
 * @return returns copy of OpenSSL X509 struct pointer that is being wrapped.
 *         NB! This struct must be freed using X509_free() function from OpenSSL or X509_scope(X509**) macro
 * @throws IOException throws exception if the X509 cert structure copy fails.
 */
X509* digidoc::X509Cert::getX509() const throw(IOException)
{
    return copyX509(cert);
}

/**
 * Creates a copy of the X509 certificate.
 *
 * @param cert X509 certificate to be copied.
 * @return returns copy of X509.
 * @throws IOException throws exception if the X509 cert structure copy fails.
 */
X509* digidoc::X509Cert::copyX509(X509* cert) throw(IOException)
{
    X509* copy = X509_dup(cert);
    if(copy == NULL)
    {
        THROW_IOEXCEPTION("Failed to copy X509 certificate: %s", ERR_reason_error_string(ERR_get_error()));
    }

    return copy;
}

/**
 * Parses X.509 PEM formatted certificate from file.
 * NB! This struct must be freed using X509_free() function from OpenSSL or
 * with X509_scope struct.
 *
 * @param path PEM formatted X.509 certificate file path.
 * @return returns certificate parsed from file.
 * @throws IOException throws exception if the file does not contain X.509
 *         PEM formatted certificate.
 */
X509* digidoc::X509Cert::loadX509(const std::string& path) throw(IOException)
{
    // Initialize OpenSSL file.
    BIO* file = BIO_new_file(path.c_str(), "rb"); BIO_scope fileScope(&file);
    if(file == NULL)
    {
        THROW_IOEXCEPTION("Failed to open X.509 certificate file '%s': %s",
                path.c_str(), ERR_reason_error_string(ERR_get_error()));
    }

    // Parse X.509 certificate from file.
    X509* cert = PEM_read_bio_X509(file, NULL, NULL, NULL);
    if(cert == NULL)
    {
        THROW_IOEXCEPTION("Failed to load X.509 certificate from file '%s': %s",
                path.c_str(), ERR_reason_error_string(ERR_get_error()));
    }

    return cert;
}

/**
 * Parses X.509 PEM formatted certificates from file. The file can contain multiple
 * certificates. You can just use 'cat' command in linux to add multiple PEM formatted
 * X.509 certificates to one file (e.g. <code>cat cert1.pem cert2.pem > certs.pem</code>).
 * NB! This struct must be freed using sk_X509_free() function from OpenSSL or
 * with X509Stack_scope struct.
 *
 * @param path PEM formatted X.509 certificates file path.
 * @return returns stack of parsed X.509 certificates.
 * @throws IOException throws exception if the file does not contain X.509
 *         PEM formatted certificate(s).
 */
STACK_OF(X509)* digidoc::X509Cert::loadX509Stack(const std::string& path) throw(IOException)
{
    // Create an empty X.509 stack.
    STACK_OF(X509)* stack = sk_X509_new_null();
    if(stack == NULL)
    {
        THROW_IOEXCEPTION("Failed to create X.509 certificate stack.");
    }

    // Initialize OpenSSL file.
    BIO* file = BIO_new_file(path.c_str(), "rb"); BIO_scope fileScope(&file);
    if(file == NULL)
    {
        THROW_IOEXCEPTION("Failed to open X.509 certificates file '%s': %s",
                path.c_str(), ERR_reason_error_string(ERR_get_error()));
    }

    // Read certificates info from the file.
    STACK_OF(X509_INFO)* certsInfo = PEM_X509_INFO_read_bio(file, NULL, NULL, NULL);
    if(certsInfo == NULL)
    {
        THROW_IOEXCEPTION("Failed to read X.509 certificates from file '%s': %s",
                path.c_str(), ERR_reason_error_string(ERR_get_error()));
    }

    // Put all found certificates to the stack.
    for(int i = 0; i < sk_X509_INFO_num(certsInfo); i++)
    {
        X509_INFO* xi = sk_X509_INFO_value(certsInfo, i);
        if(xi->x509 != NULL)
        {
            sk_X509_push(stack, xi->x509);
            xi->x509 = NULL;
        }
    }

    // Release resources.
    sk_X509_INFO_pop_free(certsInfo, X509_INFO_free);

    return stack;
}

/**
 * Encodes the X509 certificate using DER encoding.
 *
 * @return returns X509 certificate encoded in DER encoding.
 * @throws IOException throws exception if encoding failed.
 */
std::vector<unsigned char> digidoc::X509Cert::encodeDER() const throw(IOException)
{
    // Get size of the DER encodes certificate.
    int bufSize = i2d_X509(cert, NULL);
    if(bufSize < 0)
    {
        THROW_IOEXCEPTION("Failed to encode X509 cert to DER.");
    }

    // Allocate memory for the DER encoding.
    std::vector<unsigned char> derEncodedX509(bufSize, 0);

    // DER encode X.509 certificate.
    unsigned char* pBuf = &derEncodedX509[0];
    bufSize = i2d_X509(cert, &pBuf);
    if(bufSize < 0)
    {
        THROW_IOEXCEPTION("Failed to encode X509 cert to DER.");
    }

    return derEncodedX509;
}

/**
 * @return returns padding size.
 * @throws IOException exception is thrown when failed to get padding size.
 */
int digidoc::X509Cert::getPaddingSize() const throw(IOException)
{
    EVP_PKEY *pubKey = getPublicKey();
    RSA *rsa = EVP_PKEY_get1_RSA(pubKey);
    if(!rsa)
    {
        EVP_PKEY_free(pubKey);
        THROW_IOEXCEPTION("Failed to read certificate RSA key: %s", ERR_reason_error_string(ERR_get_error()));
    }
    int size = RSA_size(rsa);
    RSA_free(rsa);
    EVP_PKEY_free(pubKey);
    return size;
}

/**
 * @return returns X.509 certificate serial number.
 * @throws IOException exception is thrown if the serial is incorrect.
 */
std::string digidoc::X509Cert::getSerial() const throw(IOException)
{
    std::string serial;
    BIGNUM *bn = ASN1_INTEGER_to_BN(X509_get_serialNumber(cert), 0);
    if(bn)
    {
        char *str = BN_bn2dec(bn);
        if(str)
            serial = str;
        OPENSSL_free(str);
        BN_free(bn);
    }

    if(serial.empty())
    {
        THROW_IOEXCEPTION("Failed to read certificate serial number from X.509 certificate: %s", ERR_reason_error_string(ERR_get_error()));
    }

    return serial;
}

/**
 * @return returns X.509 certificate issuer name. Can also return NULL pointer
 *         if the issuer name is missing.
 */
X509_NAME* digidoc::X509Cert::getIssuerNameAsn1() const
{
    return X509_get_issuer_name(cert);
}

/**
 * Converts X.509 issuer name to string.
 *
 * @return issuer name converted to string.
 * @throws IOException exception is throws if the conversion failed.
 */
std::string digidoc::X509Cert::getIssuerName() const throw(IOException)
{
    X509_NAME* issuerName = X509_get_issuer_name(cert);
    if(issuerName == NULL)
    {
        THROW_IOEXCEPTION("Failed to convert X.509 certificate issuer name: %s", ERR_reason_error_string(ERR_get_error()));
    }

    return toString(issuerName);
}

/**
 * Returns current certificate key usage bits
 *
 * @return key usage bits
 */
std::vector<digidoc::X509Cert::KeyUsage> digidoc::X509Cert::getKeyUsage() const throw(IOException)
{
    ASN1_BIT_STRING *keyusage = (ASN1_BIT_STRING*)X509_get_ext_d2i(cert, NID_key_usage, 0, 0);
    if(!keyusage)
        return std::vector<KeyUsage>();

    std::vector<KeyUsage> usage;
    for(int n = 0; n < 9; ++n)
    {
        if(ASN1_BIT_STRING_get_bit(keyusage, n))
            usage.push_back(KeyUsage(n));
    }
    ASN1_BIT_STRING_free(keyusage);
    return usage;
}

/**
 * Returns current certificate policies
 *
 * @return certificate policies
 */
std::vector<std::string> digidoc::X509Cert::getCertificatePolicies() const throw(IOException)
{
    CERTIFICATEPOLICIES *cp = (CERTIFICATEPOLICIES*)X509_get_ext_d2i(cert, NID_certificate_policies, 0, 0);
    if(!cp)
        return std::vector<std::string>();

    char buf[50];
    std::vector<std::string> pol;
    for(int i = 0; i < sk_POLICYINFO_num(cp); ++i)
    {
        memset(buf, 0, 50);
        int len = OBJ_obj2txt(buf, 50, sk_POLICYINFO_value(cp, i)->policyid, 1);
        if(len != NID_undef)
            pol.push_back(std::string(buf, len));
    }
    sk_POLICYINFO_pop_free(cp, POLICYINFO_free);
    return pol;
}

/**
 * Converts X.509 subject to string.
 *
 * @return issuer name converted to string.
 * @throws IOException exception is throws if the conversion failed.
 */
std::string digidoc::X509Cert::getSubject() const throw(IOException)
{
    X509_NAME* subject = X509_get_subject_name(cert);
    if(!subject)
       THROW_IOEXCEPTION("Failed to convert X.509 certificate subject: %s", ERR_reason_error_string(ERR_get_error()));

    return toString(subject);
}

/**
 * Converts X509_NAME struct to string.
 *
 * @param name X509_NAME struct that is converted to string.
 * @return converted value of X509_NAME.
 * @throws IOException throws exception if conversion failed.
 */
std::string digidoc::X509Cert::toString(X509_NAME* name) const throw(IOException)
{
    BIO* mem = BIO_new(BIO_s_mem()); BIO_scope memScope(&mem);
    if(!mem)
    {
        THROW_IOEXCEPTION("Failed to allocate memory for X509_NAME conversion: %s", ERR_reason_error_string(ERR_get_error()));
    }

    // Convert the X509_NAME struct to string.
    if(X509_NAME_print_ex(mem, name, 0, XN_FLAG_RFC2253) < 0)
    {
        THROW_IOEXCEPTION("Failed to convert X509_NAME struct to string: %s", ERR_reason_error_string(ERR_get_error()));
    }

    // Read the converted string from buffer.
    char buf[128];
    int bytesRead;
    std::string str;
    while((bytesRead = BIO_gets(mem, &buf[0], sizeof(buf))) > 0)
    {
        str.append(buf);
    }

    return str;
}

/**
 * @return returns X.509 certificates public key.
 * @throws IOException throws exception if public key is missing.
 */
EVP_PKEY* digidoc::X509Cert::getPublicKey() const throw(IOException)
{
    EVP_PKEY* pubKey = X509_get_pubkey(cert);
    if((pubKey == NULL) || (pubKey->type != EVP_PKEY_RSA))
    {
        EVP_PKEY_free(pubKey);
        THROW_IOEXCEPTION("Unable to load RSA public Key: %s", ERR_reason_error_string(ERR_get_error()));
    }
    return pubKey;
}

/**
 * @return Extracts RSA modulus from X.509 certificate and returns it.
 * @throws IOException throws exception if the RSA modulus extraction failed.
 */
std::vector<unsigned char> digidoc::X509Cert::getRsaModulus() const throw(IOException)
{
    EVP_PKEY* pubKey = getPublicKey();

    // Get size of the RSA modulus.
    int bufSize = BN_num_bytes(pubKey->pkey.rsa->n);

    if(bufSize <= 0)
    {
        EVP_PKEY_free(pubKey);
        THROW_IOEXCEPTION("Failed to extract RSA modulus.");
    }

    // Allocate memory for the RSA modulus.
    std::vector<unsigned char> rsaModulus(bufSize, 0);

    // Extract RSA modulus.
    if(BN_bn2bin(pubKey->pkey.rsa->n, &rsaModulus[0]) <= 0)
    {
        EVP_PKEY_free(pubKey);
        THROW_IOEXCEPTION("Failed to extract RSA modulus.");
    }

    EVP_PKEY_free(pubKey);
    return rsaModulus;
}

/**
 * @return Extracts RSA exponent from X.509 certificate and returns it.
 * @throws IOException throws exception if the RSA exponent extraction failed.
 */
std::vector<unsigned char> digidoc::X509Cert::getRsaExponent() const throw(IOException)
{
    EVP_PKEY* pubKey = getPublicKey();

    // Get size of the RSA exponent.
    int bufSize = BN_num_bytes(pubKey->pkey.rsa->e);

    if(bufSize <= 0)
    {
        EVP_PKEY_free(pubKey);
        THROW_IOEXCEPTION("Failed to extract RSA exponent.");
    }

    // Allocate memory for the RSA exponent.
    std::vector<unsigned char> rsaExponent(bufSize, 0);

    // Extract RSA exponent.
    if(BN_bn2bin(pubKey->pkey.rsa->e, &rsaExponent[0]) <= 0)
    {
        EVP_PKEY_free(pubKey);
        THROW_IOEXCEPTION("Failed to extract RSA exponent.");
    }

    EVP_PKEY_free(pubKey);
    return rsaExponent;
}

bool digidoc::X509Cert::isValid() const throw(IOException)
{
	int notBefore = X509_cmp_current_time(cert->cert_info->validity->notBefore);
	int notAfter = X509_cmp_current_time(cert->cert_info->validity->notAfter);
	if(notBefore == 0 || notAfter == 0)
		THROW_IOEXCEPTION("Failed to validate cert",ERR_reason_error_string(ERR_get_error()));
	return notBefore < 0 && notAfter > 0;
}

/**
 * Check if X509Cert is signed by trusted issuer
 * @return 0 or openssl error_code. Get human readable cause with X509_verify_cert_error_string(code)
 * @throw IOException if error
 */
int digidoc::X509Cert::verify() const throw(IOException)
{
    X509_STORE_CTX *csc = X509_STORE_CTX_new(); X509_STORE_CTX_scope csct(&csc);
    if (!csc)
        THROW_IOEXCEPTION("Failed to create X509_STORE_CTX %s",ERR_reason_error_string(ERR_get_error()));

    X509_STORE *store = digidoc::X509CertStore::getInstance()->getCertStore();
    X509* x = getX509(); X509_scope xt(&x);
    if(!X509_STORE_CTX_init(csc, store, x, NULL))
        THROW_IOEXCEPTION("Failed to init X509_STORE_CTX %s",ERR_reason_error_string(ERR_get_error()));

    int ok = X509_verify_cert(csc);

    if(!ok)
    {
        int err = X509_STORE_CTX_get_error(csc);
        X509Cert cause(X509_STORE_CTX_get_current_cert (csc));
        std::ostringstream s;
        s << "Unable to verify " << cause.getSubject();
        s << ". Cause: " << X509_verify_cert_error_string(err);
        switch(err)
        {
        case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
        {
            IOException e(__FILE__, __LINE__, s.str());
            e.setCode( Exception::CertificateIssuerMissing );
            throw e;
            break;
        }
        default: THROW_IOEXCEPTION(s.str().c_str()); break;
        }
    }

    return ok;
}

/**
 * Check if X509Cert issuer is same as provided issuer name by
 * http://www.w3.org/TR/xmldsig-core/#dname-encrules which refers to
 * http://www.ietf.org/rfc/rfc4514.txt
 *
 * String X.500 AttributeType
 * CN commonName (2.5.4.3)
 * L localityName (2.5.4.7)
 * ST stateOrProvinceName (2.5.4.8)
 * O organizationName (2.5.4.10)
 * OU organizationalUnitName (2.5.4.11)
 * C countryName (2.5.4.6)
 * STREET streetAddress (2.5.4.9)
 * DC domainComponent (0.9.2342.19200300.100.1.25)
 * UID userId (0.9.2342.19200300.100.1.1)
 *
 * These attribute types are described in [RFC4519].
 * Implementations MAY recognize other DN string representations.
 * However, as there is no requirement that alternative DN string
 * representations be recognized (and, if so, how), implementations
 * SHOULD only generate DN strings in accordance with Section 2 of this document.
 *
 * @param issuer name
 * @return 0 if equal, otherwise a number different from 0 is returned
 * @throw IOException if error
 */
int digidoc::X509Cert::compareIssuerToString(const std::string &name) const throw(IOException)
{
    size_t pos = 0, old = 0;
    while(true)
    {
        pos = name.find(",", old);
        if(pos == std::string::npos)
        {
            pos = name.size();
            if(pos < old)
                break;
        }
        else
        {
            if(name.compare(pos-1, 1, "\\") == 0)
                continue;
        }

        std::string nameitem = name.substr(old, pos - old);
        old = pos + 1;

        if((pos = nameitem.find("=")) == std::string::npos ||
            nameitem.compare(pos-1, 1, "\\") == 0)
            continue;

        std::string obj = nameitem.substr(0, pos);
        if(obj.compare("CN") != 0 && obj.compare("commonName") != 0 &&
           obj.compare("L") != 0 && obj.compare("localityName") != 0 &&
           obj.compare("ST") != 0 && obj.compare("stateOrProvinceName") != 0 &&
           obj.compare("O") != 0 && obj.compare("organizationName") != 0 &&
           obj.compare("OU") != 0 && obj.compare("organizationalUnitName") != 0 &&
           obj.compare("C" ) != 0 && obj.compare("countryName") != 0 &&
           obj.compare("STREET") != 0 && obj.compare("streetAddress") != 0 &&
           obj.compare("DC") != 0 && obj.compare("domainComponent") != 0 &&
           obj.compare("UID") != 0 && obj.compare("userId") != 0)
            continue;

        X509_NAME_ENTRY *enta = X509_NAME_ENTRY_create_by_txt(0, obj.c_str(),
            MBSTRING_UTF8, (unsigned char*)nameitem.substr(pos+1, pos-old).c_str(), -1);
        if(!enta)
            return -1;

        ASN1_OBJECT *obja = X509_NAME_ENTRY_get_object(enta);

        bool found = false;
        for(int i = 0; i < X509_NAME_entry_count(getIssuerNameAsn1()); ++i)
        {
            X509_NAME_ENTRY *entb = X509_NAME_get_entry(getIssuerNameAsn1(), i);
            ASN1_OBJECT *objb = X509_NAME_ENTRY_get_object(entb);
            if(memcmp(obja->data, objb->data, obja->length) == 0 &&
                memcmp(enta->value, entb->value, enta->size) == 0)
            {
                found = true;
                break;
            }
        }
        X509_NAME_ENTRY_free(enta);
        if(!found)
            return -1;
    }
    return 0;
}
