#include "X509CertStore_p.h"

#if defined(_WIN32) && USE_NATIVE_CERTSTORE
#include "MSX509CertStore.h"
#elif defined(__APPLE__) && USE_NATIVE_CERTSTORE
#include "MACX509CertStore.h"
#endif
#include "DirectoryX509CertStore.h"
#include "../../Conf.h"
#include "../../log.h"

using namespace digidoc;

int X509CertStorePrivate::verify_callback(int ok, X509_STORE_CTX *ctx)
{
    switch(X509_STORE_CTX_get_error(ctx))
    {
    case X509_V_ERR_CERT_HAS_EXPIRED: return 1;
    default: return ok;
    }
}

/**
 * X.509 certificate store implementation.
 */
X509CertStore* X509CertStore::INSTANCE = 0;

/**
 * X509CertStore constructor.
 */
X509CertStore::X509CertStore()
    : d(new X509CertStorePrivate)
{
    d->stack = sk_X509_new_null();
    d->store = X509_STORE_new();
    X509_STORE_set_verify_cb_func(d->store, d->verify_callback);
}

/**
 * Release all certificates.
 */
X509CertStore::~X509CertStore()
{
    X509_STORE_free(d->store);
    sk_X509_pop_free(d->stack, X509_free);
    delete d;
}

/**
 * Sets the X.509 certificate store implementation.
 *
 * @param impl X.509 certificate store implementation or 0 for platform default.
 */
void X509CertStore::init(X509CertStore *impl)
{
    delete INSTANCE;
    if(!impl)
    {
        std::string path = Conf::getInstance()->getCertStorePath();
        if(path.empty())
#if defined(_WIN32) && USE_NATIVE_CERTSTORE
            INSTANCE = new digidoc::MSX509CertStore();
#elif defined(__APPLE__) && USE_NATIVE_CERTSTORE
            INSTANCE = new digidoc::MACX509CertStore();
#else
            INSTANCE = new digidoc::DirectoryX509CertStore( CERTS_PATH );
#endif
        else
            INSTANCE = new digidoc::DirectoryX509CertStore( path );
    }
    else
        INSTANCE = impl;
}

/**
 * Releases the X.509 certificate store implementation.
 */
void X509CertStore::destroy()
{
    delete INSTANCE;
    INSTANCE = 0;
}

/**
 * @return returns the X.509 certificate store implementation.
 */
X509CertStore* X509CertStore::getInstance() throw(IOException)
{
    if (!INSTANCE)
        THROW_IOEXCEPTION("X509CertStore is not initialized");
    return INSTANCE;
}


/**
 * Return STACK_OF(X509) containing all certs loaded from directory
 * @return STACK_OF(X509) all certs in store.
 * throws IOException
 */
STACK_OF(X509)* digidoc::X509CertStore::getCerts() const  throw(IOException)
{
    return d->stack;
}

/**
 * Return X509_STORE containing all certs loaded from directory
 * @return X509_STORE all certs in store.
 * throws IOException
 */
X509_STORE* digidoc::X509CertStore::getCertStore() const throw(IOException)
{
    return d->store;
}

/**
 * Searches certificate by serial number and returns a copy of it if found.
 * If not found returns <code>NULL</code>.
 * NB! The returned certificate must be freed with OpenSSL function X509_free(X509* cert)
 * or X509_scope(X509**) macro.
 *
 * @param certSerial certificate serial number.
 * @return returns copy of found certificate or <code>NULL</code> if certificate was not found.
 * @throws IOException exception is thrown if copying certificate failed.
 */
X509* digidoc::X509CertStore::getCert(long certSerial) const throw(IOException)
{
    for(int i = 0; i < sk_X509_num(d->stack); ++i)
    {
        X509 *x = sk_X509_value(d->stack, i);
        if(ASN1_INTEGER_get(X509_get_serialNumber(x)) == certSerial)
            return X509Cert::copyX509(x);
    }
    return 0;
}

/**
 * Searches certificate by subject and returns a copy of it if found.
 * If not found returns <code>NULL</code>.
 * NB! The returned certificate must be freed with OpenSSL function X509_free(X509* cert).
 *
 * @param subject certificate subject.
 * @return returns copy of found certificate or <code>NULL</code> if certificate was not found.
 * @throws IOException exception is thrown if copying certificate failed.
 */
X509* digidoc::X509CertStore::getCert(const X509_NAME &subject) const throw(IOException)
{
    for(int i = 0; i < sk_X509_num(d->stack); ++i)
    {
        X509 *x = sk_X509_value(d->stack, i);
        if(X509_NAME_cmp(X509_get_subject_name(x), &subject) == 0)
            return X509Cert::copyX509(x);
    }
    return 0;
}

/**
 * Searches certificates by object type and returns a stack.
 * NB! The returned certificate must be freed with OpenSSL function X509_free(X509* cert).
 *
 * @param subject certificate object.
 * @param subject certificate name.
 * @return returns copy of found certificates
 * @throws IOException exception is thrown if copying certificate failed.
 */
STACK_OF(X509)* digidoc::X509CertStore::findCerts(const std::string &obj, const std::string &name) const throw(IOException)
{
    STACK_OF(X509) *result = sk_X509_new_null();
    if(!result)
        THROW_IOEXCEPTION("Failed to create X.509 certificate stack.");

    for(int i = 0; i < sk_X509_num(d->stack); ++i)
    {
        X509 *x = sk_X509_value(d->stack, i);
        if(X509_cmp_current_time(x->cert_info->validity->notBefore) >= 0 &&
           X509_cmp_current_time(x->cert_info->validity->notAfter) <= 0)
            continue;

        X509_NAME *n = X509_get_subject_name(x);
        for(int i = 0; i < X509_NAME_entry_count(n); ++i)
        {
            X509_NAME_ENTRY *e = X509_NAME_get_entry(n, i);

            const char *o = OBJ_nid2sn(OBJ_obj2nid(X509_NAME_ENTRY_get_object(e)));
            if(obj != o)
                continue;

            char *data = 0;
            int size = ASN1_STRING_to_UTF8((unsigned char**)&data, X509_NAME_ENTRY_get_data(e));
            if(name == std::string(data, size))
                sk_X509_push(result, X509_dup(x));
            OPENSSL_free(data);
        }
    }

    return result;
}
