#include "EstEIDSigner.h"

#include "../Digest.h"
#include "../cert/X509Cert.h"
#include "../../Conf.h"
#include "../../log.h"

#include <algorithm>
#include <string.h>

/**
 * Initializes base class PKCS11Signer.
 *
 * @param driver full path to the PKCS #11 driver (e.g. /usr/lib/opensc-pkcs11.so)
 * @throws SignException exception is thrown if the provided PKCS #11 driver
 *         loading failed.
 */
digidoc::EstEIDSigner::EstEIDSigner(const std::string& driver) throw(SignException)
 : PKCS11Signer(driver)
{
}

/**
 *
 */
digidoc::EstEIDSigner::~EstEIDSigner()
{
}

/**
 * Implements signing certificate selector for EstEID ID-Cards.
 *
 * @param certificates list of certificates to choose from. List of all certificates
 *        found ID-card.
 * @return should return the selected certificate.
 * @throws SignException throws exception if no suitable certificate was found.
 */
digidoc::PKCS11Signer::PKCS11Cert digidoc::EstEIDSigner::selectSigningCertificate(
        const std::vector<PKCS11Signer::PKCS11Cert> &certificates) const throw(SignException)
{
    if(certificates.empty())
        THROW_SIGNEXCEPTION("Could not find certificate.");
    for(std::vector<PKCS11Signer::PKCS11Cert>::const_iterator i = certificates.begin(); i < certificates.end(); ++i)
    {
        std::vector<digidoc::X509Cert::KeyUsage> usage = digidoc::X509Cert(i->cert).getKeyUsage();
        if( find( usage.begin(), usage.end(), digidoc::X509Cert::NonRepudiation ) != usage.end() )
            return *i;
    }
    THROW_SIGNEXCEPTION("Could not find certificate.");
    return PKCS11Signer::PKCS11Cert();
}

int digidoc::EstEIDSigner::type() const
{
    int result = digidoc::Digest::toMethod( Conf::getInstance()->getSignatureUri() );
    if( result == NID_sha1 )
        return result;
    bool found = false;
    std::vector<std::string> pol = digidoc::X509Cert( getCert() ).getCertificatePolicies();
    for( std::vector<std::string>::iterator i = pol.begin(); i != pol.end(); ++i )
    {
        if(i->find("1.3.6.1.4.1.10015.1.2.", 22) == 0 ||
           i->find("1.3.6.1.4.1.10015.3.2.", 22) == 0)
            found = true;
    }
    if(!found)
        return X509Cert(getCert()).getPaddingSize() > 128 ? result : NID_sha224;
    return result;
}
