#include "PKCS11Signer.h"

#include "pkcs11.h"

#include "../../log.h"
#include "../../Conf.h"
#include "../../util/File.h"
#include "../cert/X509Cert.h"

#include <sstream>
#include <string.h>
#ifdef _WIN32
#include <Windows.h>
#else
#include <dlfcn.h>
#endif

namespace digidoc
{

struct SignSlot
{
    X509* certificate;
    CK_SLOT_ID slot;
    CK_ULONG cert;
};

class PKCS11SignerPrivate
{
public:
    PKCS11SignerPrivate()
    : f(0)
    , slots(0)
    , numberOfSlots(0)
    {
        sign.certificate = 0;
        sign.slot = 0;
        sign.cert = 0;
    };

    PKCS11Signer::PKCS11Cert createPKCS11Cert(CK_TOKEN_INFO *slot, X509 *cert);
    bool attribute( CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
        CK_ATTRIBUTE_TYPE type, CK_VOID_PTR value, CK_ULONG &size );
    std::vector<CK_OBJECT_HANDLE> findObject( CK_SESSION_HANDLE session, CK_OBJECT_CLASS cls );

#ifdef _WIN32
    bool load(const std::string &driver)
    {
        std::wstring _driver = util::File::encodeName(driver);
        return (h = LoadLibraryW(_driver.c_str())) != 0;
    }

    void* resolve(const char *symbol)
    { return h ? (void*)GetProcAddress(h, symbol) : 0; }

    void unload()
    { if(h) FreeLibrary(h); h = 0; }
#else
    bool load(const std::string &driver)
    { return (h = dlopen(driver.c_str(), RTLD_LAZY)); }

    void* resolve(const char *symbol)
    { return h ? dlsym(h, symbol) : 0; }

    void unload()
    { if(h) dlclose(h); h = 0; }
#endif

    CK_FUNCTION_LIST *f;
#ifdef _WIN32
    HINSTANCE h;
#else
    void *h;
#endif

    SignSlot sign;
    CK_SLOT_ID_PTR slots;
    CK_ULONG numberOfSlots;

    static const unsigned char sha1[];
    static const unsigned char sha224[];
    static const unsigned char sha256[];
    static const unsigned char sha384[];
    static const unsigned char sha512[];
};

}

using namespace digidoc;

const unsigned char PKCS11SignerPrivate::sha1[] =
{ 0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05, 0x00, 0x04, 0x14 };

const unsigned char PKCS11SignerPrivate::sha224[] =
{ 0x30, 0x2d, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x04, 0x05, 0x00, 0x04, 0x1c };

const unsigned char PKCS11SignerPrivate::sha256[] =
{ 0x30, 0x31, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0x04, 0x20 };

const unsigned char PKCS11SignerPrivate::sha384[] =
{ 0x30, 0x41, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x02, 0x05, 0x00, 0x04, 0x30 };

const unsigned char PKCS11SignerPrivate::sha512[] =
{ 0x30, 0x51, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03, 0x05, 0x00, 0x04, 0x40 };



bool PKCS11SignerPrivate::attribute( CK_SESSION_HANDLE session,
    CK_OBJECT_HANDLE obj, CK_ATTRIBUTE_TYPE type, CK_VOID_PTR value, CK_ULONG &size )
{
    CK_ATTRIBUTE attr = { type, value, size };
    CK_RV err = f->C_GetAttributeValue( session, obj, &attr, 1 );
    size = attr.ulValueLen;
    return err == CKR_OK;
}

std::vector<CK_OBJECT_HANDLE> PKCS11SignerPrivate::findObject( CK_SESSION_HANDLE session, CK_OBJECT_CLASS cls )
{
    CK_ATTRIBUTE attr = { CKA_CLASS, &cls, sizeof(cls) };
    if( f->C_FindObjectsInit( session, &attr, 1 ) != CKR_OK )
        return std::vector<CK_OBJECT_HANDLE>();

    CK_ULONG count = 32;
    std::vector<CK_OBJECT_HANDLE> result( count );
    CK_RV err = f->C_FindObjects( session, &result[0], result.size(), &count );
    result.resize( err == CKR_OK ? count : 0 );
    f->C_FindObjectsFinal( session );
    return result;
}

/**
 * Initializes p11 library and loads PKCS #11 driver. Get driver path
 * from configuration.
 *
 * @throws SignException exception is thrown if the provided PKCS #11 driver
 *         loading failed.
 */
PKCS11Signer::PKCS11Signer() throw(SignException)
 : d(new PKCS11SignerPrivate)
{
    loadDriver(Conf::getInstance()->getPKCS11DriverPath());
}

/**
 * Initializes p11 library and loads PKCS #11 driver.
 *
 * @param driver full path to the PKCS #11 driver (e.g. /usr/lib/opensc-pkcs11.so)
 * @throws SignException exception is thrown if the provided PKCS #11 driver
 *         loading failed.
 */
PKCS11Signer::PKCS11Signer(const std::string &driver) throw(SignException)
 : d(new PKCS11SignerPrivate)
{
    loadDriver(driver);
}

/**
 * Uninitializes p11 library and releases acquired memory.
 */
PKCS11Signer::~PKCS11Signer()
{
    DEBUG("~PKCS11Signer()");
    unloadDriver();
	delete d;
}

void PKCS11Signer::unloadDriver()
{
    if(!d->f)
        return;

    if(!d->slots)
    {
        // Release all slots.
        delete [] d->slots;
        d->slots = 0;
        d->numberOfSlots = 0;
    }

    // Release PKCS #11 context.
    d->f->C_Finalize( 0 );
    d->f = 0;
    d->unload();
}

/**
 * Initializes p11 library and loads PKCS #11 driver.
 *
 * @param driver full path to the PKCS #11 driver (e.g. /usr/lib/opensc-pkcs11.so)
 * @throws SignException exception is thrown if the provided PKCS #11 driver
 *         loading failed.
 */
void PKCS11Signer::loadDriver(const std::string &driver) throw(SignException)
{
    DEBUG("PKCS11Signer(driver = '%s')", driver.c_str());
    if(driver.empty())
        THROW_SIGNEXCEPTION("Failed to load driver for PKCS #11 engine: %s.", driver.c_str());

    unloadDriver();

    if(!d->load(driver))
        THROW_SIGNEXCEPTION("Failed to load driver for PKCS #11 engine: %s.", driver.c_str());

    CK_C_GetFunctionList l = CK_C_GetFunctionList(d->resolve( "C_GetFunctionList" ));
    if( !l ||
        l( &d->f ) != CKR_OK ||
        d->f->C_Initialize( 0 ) != CKR_OK )
        THROW_SIGNEXCEPTION("Failed to load driver for PKCS #11 engine: %s.", driver.c_str());
}

void PKCS11Signer::loadDriver() throw(SignException)
{
    if(Conf::getInstance())
        loadDriver(Conf::getInstance()->getPKCS11DriverPath());
}

/**
 * Finds all slots connected with the computer, if the slots have tokens, lists all
 * certificates found in token. If there are more that 1 certificate lets the user
 * application select (by calling the <code>selectSignCertificate</code> callback
 * function) the certificate used for signing.
 *
 * @return returns certificate used for signing.
 * @throws throws exception if failed to select the signing certificate. For example
 *         no cards found or card has no certificate.
 */
X509* PKCS11Signer::getCert() const throw(SignException)
{
    DEBUG("PKCS11Signer::getCert()");

    // If certificate is already selected return it.
    if(d->sign.certificate)
        return d->sign.certificate;

    // Set selected state to 'no certificate selected'.
    d->sign.certificate = 0;
    d->sign.slot = 0;
    if(d->slots)
    {
        // Release all slots.
        delete [] d->slots;
        d->slots = 0;
        d->numberOfSlots = 0;
    }

    // Load all slots.
    if(d->f->C_GetSlotList(true, 0, &d->numberOfSlots) != CKR_OK)
        THROW_SIGNEXCEPTION("Could not find any ID-Cards in any readers");
    d->slots = new CK_SLOT_ID[d->numberOfSlots];
    if(d->f->C_GetSlotList(true, d->slots, &d->numberOfSlots) != CKR_OK)
        THROW_SIGNEXCEPTION("Could not find any ID-Cards in any readers");

    // Iterate over all found slots, if the slot has a token, check if the token has any certificates.
    CK_SESSION_HANDLE session = 0;
    std::vector<PKCS11Signer::PKCS11Cert> certificates;
    std::vector<SignSlot> certSlotMapping;
    for(CK_ULONG i = 0; i < d->numberOfSlots; ++i)
    {
        CK_TOKEN_INFO token;
        std::vector<CK_OBJECT_HANDLE> objs;

        if(session)
           d->f->C_CloseSession(session);

        if(d->f->C_GetTokenInfo(d->slots[i], &token) != CKR_OK ||
           d->f->C_OpenSession(d->slots[i], CKF_SERIAL_SESSION, 0, 0, &session) != CKR_OK ||
           (objs = d->findObject(session, CKO_CERTIFICATE)).empty())
            continue;

        for(CK_ULONG j = 0; j < objs.size(); ++j)
        {
            CK_ULONG size = 0;
            if(!d->attribute(session, objs[j], CKA_VALUE, 0, size))
                continue;
            unsigned char *value = new unsigned char[size];
            if(!d->attribute(session, objs[j], CKA_VALUE, value, size))
                continue;

            X509 *x509 = 0;
            const unsigned char *p = value;
            d2i_X509(&x509, &p, size);
            delete [] value;

            if(!X509Cert(x509).isValid())
            {
                X509_free(x509);
                continue;
            }

            SignSlot signSlot = { x509, d->slots[i], j };
            certSlotMapping.push_back(signSlot);
            certificates.push_back(d->createPKCS11Cert(&token, x509));
        }
    }
    if(session)
        d->f->C_CloseSession(session);

    if(certificates.empty())
        THROW_SIGNEXCEPTION("No certificates found.");

    // Let the application select the signing certificate.
    X509* selectedCert = selectSigningCertificate(certificates).cert;
    if(!selectedCert)
        THROW_SIGNEXCEPTION("No certificate selected.");

    // Find the corresponding slot and PKCS11 certificate struct.
    for(std::vector<SignSlot>::const_iterator i = certSlotMapping.begin(); i != certSlotMapping.end(); ++i)
    {
        if(!d->sign.cert && X509_cmp(i->certificate, selectedCert) == 0)
            d->sign = *i;
        else
            X509_free(i->certificate);
    }

    if(!d->sign.certificate)
        THROW_SIGNEXCEPTION("Could not find slot for selected certificate.");

    return d->sign.certificate;
}

/**
 * Signs the digest provided using the selected certificate. If the certificate needs PIN,
 * the PIN is acquired by calling the callback function <code>getPin</code>.
 *
 * @param digest digest, which is being signed.
 * @param signature memory for the signature that is created. Struct parameter <code>length</code>
 *        is set to the actual signature length.
 * @throws SignException throws exception if the signing operation failed.
 */
void PKCS11Signer::sign(const Digest &digest, Signature &signature) throw(SignException)
{
    DEBUG("sign(digest = {type=%s,digest=%p,length=%d}, signature={signature=%p,length=%d})",
        OBJ_nid2sn(digest.type), digest.digest, digest.length, signature.signature, signature.length);

    // Check that sign slot and certificate are selected.
    if(!d->sign.certificate)
        THROW_SIGNEXCEPTION("Signing slot or certificate are not selected.");

    // Login if required.
    CK_TOKEN_INFO token;
    CK_SESSION_HANDLE session;
    if(d->f->C_GetTokenInfo(d->sign.slot, &token) != CKR_OK ||
       d->f->C_OpenSession(d->sign.slot, CKF_SERIAL_SESSION, 0, 0, &session) != CKR_OK)
        THROW_SIGNEXCEPTION("Signing slot or certificate are not selected.");

    if(token.flags & CKF_LOGIN_REQUIRED)
    {
        int rv = CKR_OK;
        if(token.flags & CKF_PROTECTED_AUTHENTICATION_PATH)
            rv = d->f->C_Login(session, CKU_USER, 0, 0);
        else
        {
            std::string pin = getPin(d->createPKCS11Cert(&token, d->sign.certificate));
            rv = d->f->C_Login(session, CKU_USER, CK_BYTE_PTR(pin.c_str()), CK_ULONG(pin.size()));
        }
        switch(rv)
        {
        case CKR_OK: break;
        case CKR_CANCEL:
        case CKR_FUNCTION_CANCELED:
        {
            SignException e( __FILE__, __LINE__, "PIN acquisition canceled.");
            e.setCode( Exception::PINCanceled );
            throw e;
            break;
        }
        case CKR_PIN_INCORRECT:
        {
            SignException e( __FILE__, __LINE__, "PIN Incorrect" );
            e.setCode( Exception::PINIncorrect );
            throw e;
            break;
        }
        case CKR_PIN_LOCKED:
        {
            SignException e( __FILE__, __LINE__, "PIN Locked" );
            e.setCode( Exception::PINLocked );
            throw e;
            break;
        }
        default:
            std::ostringstream s;
            s << "Failed to login to token '" << token.label << "': " << rv;
            SignException e( __FILE__, __LINE__, s.str() );
            e.setCode( Exception::PINFailed );
            throw e;
            break;
        }
    }

    std::vector<CK_OBJECT_HANDLE> key = d->findObject( session, CKO_PRIVATE_KEY );
    if( key.empty() )
        THROW_SIGNEXCEPTION("Could not get key that matches selected certificate.");

    // Sign the digest.
    CK_MECHANISM mech = { CKM_RSA_PKCS, 0, 0 };
    if( d->f->C_SignInit( session, &mech, key[d->sign.cert] ) != CKR_OK )
        THROW_SIGNEXCEPTION("Failed to sign digest");

    const unsigned char *sha = 0;
    size_t size = 0;
    switch( digest.type )
    {
    case NID_sha1: sha = PKCS11SignerPrivate::sha1; size = sizeof(PKCS11SignerPrivate::sha1); break;
    case NID_sha224: sha = PKCS11SignerPrivate::sha224; size = sizeof(PKCS11SignerPrivate::sha224); break;
    case NID_sha256: sha = PKCS11SignerPrivate::sha256; size = sizeof(PKCS11SignerPrivate::sha256); break;
    case NID_sha384: sha = PKCS11SignerPrivate::sha384; size = sizeof(PKCS11SignerPrivate::sha384); break;
    case NID_sha512: sha = PKCS11SignerPrivate::sha512; size = sizeof(PKCS11SignerPrivate::sha512); break;
    default: break;
    }
    unsigned char *data = new unsigned char[size + digest.length];
    memmove( data, sha, size );
    memmove( data + size, digest.digest, digest.length );
    size = size + digest.length;

    if(d->f->C_Sign(session, data, CK_ULONG(size), 0, CK_ULONG_PTR(&signature.length)) != CKR_OK)
    {
        delete [] data;
        THROW_SIGNEXCEPTION("Failed to sign digest");
    }

    signature.signature = new unsigned char[signature.length];
    CK_RV err = d->f->C_Sign(session, data, CK_ULONG(size), signature.signature, CK_ULONG_PTR(&signature.length));
    delete [] data;
    if(err != CKR_OK)
        THROW_SIGNEXCEPTION("Failed to sign digest");
}

/**
 * Helper method, creates PKCS11Cert struct. NB! token should not be NULL.
 *
 * @param slot slot to be used to init PKCS11Cert.
 * @param cert cert to be used to init PKCS11Cert.
 * @return returns created PKCS11Cert struct.
 */
PKCS11Signer::PKCS11Cert PKCS11SignerPrivate::createPKCS11Cert(CK_TOKEN_INFO *token, X509 *cert)
{
    PKCS11Signer::PKCS11Cert certificate;
    certificate.token.label = (char*)token->label;
    certificate.token.manufacturer = (char*)token->manufacturerID;
    certificate.token.model = (char*)token->model;
    certificate.token.serialNr = (char*)token->serialNumber;
    certificate.label = X509Cert( cert ).getSubject();
    certificate.cert = cert;
    return certificate;
}
