#include "log.h"
#include "BDoc.h"
#include "Conf.h"
#include "Document.h"
#include "Signature.h"
#include "crypto/signer/EstEIDConsolePinSigner.h"
#include "util/File.h"
#include "version.h"

#ifdef _WIN32
#include <Windows.h>
#endif

#define SUCCESS 0
#define FAILURE 1

using namespace digidoc;

const char* CERT_STATUSES[3] = { "GOOD", "REVOKED", "UNKNOWN" };

static int create(int argc, char *argv[]);
static int open(int argc, char *argv[]);
static void printUsage(const char *executable);
static int sign(int argc, char *argv[]);

static std::string decodeParameter( const std::string &param )
{
#ifdef _WIN32
    int len = MultiByteToWideChar(CP_ACP, 0, param.data(), int(param.size()), 0, 0);
    wchar_t *conv = new wchar_t[len];
    len = MultiByteToWideChar(CP_ACP, 0, param.data(), int(param.size()), conv, len);
    std::wstring out(conv, len);
    delete [] conv;
    return util::File::decodeName( out );
#else
    return util::File::decodeName( param );
#endif
}


/**
 * For demonstration purpose overwrites certificate selection to print out all
 * the certificates available on ID-Card.
 */
class DemoEstEIDConsolePinSigner : public EstEIDConsolePinSigner
{
public:
    DemoEstEIDConsolePinSigner(const std::string &driver, const std::string &pin) throw(SignException);
    virtual ~DemoEstEIDConsolePinSigner() {}

private:
    std::string getPin(const PKCS11Cert &certificate) throw(SignException);
    PKCS11Signer::PKCS11Cert selectSigningCertificate(
        const std::vector<PKCS11Signer::PKCS11Cert> &certificates) const throw(SignException);
    void printPKCS11Cert(const PKCS11Signer::PKCS11Cert &cert) const;

    std::string m_pin;
};



/**
 * Executes digidoc demonstration application.
 *
 * @param argc number of command line arguments.
 * @param argv command line arguments.
 * @return FAILURE (1) - failure, SUCCESS (0) - success
 */
int main(int argc, char* argv[])
{
    int returnCode = FAILURE;

    digidoc::initialize();

    printf("Version\n");
    printf("  digidoc-tool version: %s\n", VER_STR(FILE_VER_DOT));
    printf("  libdigidoc version: %s\n", digidoc::version().c_str());

    if(argc < 2)
    {
        printUsage(argv[0]);
        return FAILURE;
    }

    std::string command(argv[1]);

    if(command.compare("open") == 0)
        returnCode = open(argc, argv);
    else if(command.compare("create") == 0)
        returnCode = create(argc, argv);
    else if(command.compare("sign") == 0)
        returnCode = sign(argc, argv);
    else if(command.compare("version") == 0)
        returnCode = SUCCESS;
    else
        printUsage(argv[0]);

    digidoc::terminate();

    return returnCode;
}

/**
 * Prints demo application usage.
 */
void printUsage(const char *executable)
{
    printf("Usage: %s COMMAND [OPTIONS FILE]\n\n", executable);
    printf("  Command create:\n");
    printf("    Example: %s create --file=file1.txt --file=file2.txt --profile=TM demo-container.bdoc\n", executable);
    printf("    Available options:\n");
    printf("      --file         - The option can occur multiple times. File(s) to be signed\n");
    printf("      --city         - optional, city of production place\n");
    printf("      --state        - optional, state of production place\n");
    printf("      --postalCode   - optional, postalCode of production place\n");
    printf("      --country      - optional, country of production place\n");
    printf("      --role         - optional, option can occur multiple times. Signer role(s)\n");
    printf("      --pkcs11Driver - optional, default is '%s'. Path of PKCS11 driver.\n", Conf::getInstance()->getPKCS11DriverPath().c_str());
    printf("      --pin          - optional, default asks pin from prompt\n");
    printf("      --profile      - optional, default TM. Signing profile, available values: BES, TM\n");
    printf("  Command open:\n");
    printf("    Example: %s open --list --validateOnline container-file.bdoc\n", executable);
    printf("    Available options:\n");
    printf("      --extractAll      - extracts documents (to path when provided)\n");
    printf("      --list            - lists documents and signatures\n");
    printf("      --validateOnline  - Whether online validate all signatures\n");
    printf("  Command sign:\n");
    printf("    Example: %s sign --profile=TM demo-container.bdoc\n", executable);
    printf("    Available options:\n");
    printf("      --city         - optional, city of production place\n");
    printf("      --state        - optional, state of production place\n");
    printf("      --postalCode   - optional, postalCode of production place\n");
    printf("      --country      - optional, country of production place\n");
    printf("      --role         - optional, option can occur multiple times. Signer role(s)\n");
    printf("      --pkcs11Driver - optional, default is '%s'. Path of PKCS11 driver.\n", Conf::getInstance()->getPKCS11DriverPath().c_str());
    printf("      --pin          - optional, default asks pin from prompt\n");
    printf("      --profile      - optional, default TM. Signing profile, available values: BES, TM\n");
}

std::string parseException(const Exception &e)
{
    std::string result = e.getMsg() + "\n";
    for(size_t i = 0; i < e.getCauses().size(); ++i)
        result += parseException(e.getCauses()[i]);
    return result;
}

/**
 * Open BDOC container
 * Example command: <code>demo open --validateOnline container-file.bdoc</code>
 * Args:
 * <ul>
 *   <li><b>--extract</b> - extracts documents (to path when provided)
 *   <li><b>--list</b> - lists documents and signatures.
 *   <li><b>--validateOnline</b> - Whether online validate all signatures</li>
 * </ul>
 *
 * @param argc number of command line arguments.
 * @param argv command line arguments.
 * @return FAILURE (1) - failure, SUCCESS (0) - success
 */
int open(int argc, char* argv[])
{
    bool validateOnline = false, listContent = false, extractAll = false;
    std::string path, extractPath;    
    bool validStatus = true, extractStatus = true;

    // Parse command line arguments.
    for(int i = 2; i < argc; i++)
    {
        std::string arg( decodeParameter( argv[i] ) );
        if(arg.find("--validateOnline") == 0)
        {
            validateOnline = true;
        }
        else if(arg.find("--list") == 0)
        {
            listContent = true;
        }
        else if(arg.find("--extractAll") == 0)
        {
            extractAll = true;
            size_t pos = arg.find("=");
            if(pos != std::string::npos)
                extractPath = arg.substr(pos + 1);
        }
        else
            path = arg;
    }

    if(path.empty())
    {
        printUsage(argv[0]);
        return FAILURE;
    }

    try
    {
        // Open BDOC container.
        BDoc bdoc(path);

        if(extractAll)
        {
            printf("Extracting documents:\n");            
            for(unsigned int i = 0; i < bdoc.documentCount(); ++i)
            {
                try {
                    Document doc = bdoc.getDocument(i);
                    std::string dst = extractPath.empty() ? doc.getFileName() : extractPath + "/" + doc.getFileName();
                    doc.saveAs(dst);
                    printf("  Document %u (%s) extracted to %s (%lu bytes)\n", i, doc.getMediaType().c_str(), dst.c_str(), doc.getSize());
                }
                catch(...) {
                    printf("  Document %u extraction: FAILED\n", i);
                    extractStatus = false;
                }
            }
        }

        if(listContent)
        {
            // Print container document list.
            printf("Documents (%u):\n", bdoc.documentCount());
            for(unsigned int i = 0; i < bdoc.documentCount(); i++)
            {
                Document doc = bdoc.getDocument(i);
                printf("  Document %u (%s): %s (%lu bytes)\n", i, doc.getMediaType().c_str(), doc.getFileName().c_str(), doc.getSize());
            }

            // Print container signatures list.
            printf("\nSignatures (%u):\n", bdoc.signatureCount());

            for(unsigned int i = 0; i < bdoc.signatureCount(); i++)
            {
                // Get signature from bdoc cotainer.
                const Signature* sig = bdoc.getSignature(i);

                printf("  Signature %u (%s):\n", i, sig->getMediaType().c_str());

                // Validate signature offine. Checks, whether signature format is correct
                // and signed documents checksums are correct.
                try
                {
                    sig->validateOffline();
                    printf("    Offline validation: OK\n");

                }
                catch(const SignatureException& e)
                {
                    // If signature is not offline valid, other data fields can not be read from it.
                    printf("    Offline validation: FAILED (%s)\n", e.getMsg().c_str());
                    validStatus = false;
                    continue;
                }

                // Validate signature online. Checks the validity of the certificate used to sign the bdoc container.
                if(validateOnline)
                {
                    try
                    {
                        printf("    Online validation: %s\n", CERT_STATUSES[sig->validateOnline()]);
                    }
                    catch(const SignatureException& e)
                    {
                        printf("    Online validation: FAILED (%s)\n", e.getMsg().c_str());
                        validStatus = false;
                    }
                }

                // Get signature production place info.
                SignatureProductionPlace spp = sig->getProductionPlace();
                if(!spp.isEmpty())
                {
                    printf("    Signature production place:\n");
                    printf("      Country:           %s\n", spp.countryName.c_str());
                    printf("      State or Province: %s\n", spp.stateOrProvince.c_str());
                    printf("      City:              %s\n", spp.city.c_str());
                    printf("      Postal code:       %s\n", spp.postalCode.c_str());
                }

                // Get signer role info.
                SignerRole roles = sig->getSignerRole();
                if(!roles.isEmpty())
                {
                    printf("    Signer role(s):\n");
                    for(std::vector<std::string>::const_iterator iter = roles.claimedRoles.begin(); iter != roles.claimedRoles.end(); iter++)
                    {
                        printf("      %s\n", iter->c_str());
                    }
                }

                // Get signing time.
                printf("    Signing time: %s\n", sig->getSigningTime().c_str());

                // Get signer certificate.
                // TODO: method getSigningCertificate() does not work, implement cert printing after it is fixed.
                //X509Cert cert = sig->getSigningCertificate();
            }

        }
    }
    catch(const BDocException& e)
    {
        printf("Caught BDocException: %s\n", parseException(e).c_str());
    }
    catch(const IOException& e)
    {
        printf("Caught IOException: %s\n", parseException(e).c_str());
    }
    catch(const SignException& e)
    {
        printf("Caught SignException: %s\n", parseException(e).c_str());
    }
    catch(const SignatureException& e)
    {
        printf("Caught SignatureException: %s\n", parseException(e).c_str());
    }
    catch(const Exception& e)
    {
        printf("Caught Exception: %s\n", parseException(e).c_str());
    }
    catch(...)
    {
        printf("Caught unknown exception\n");
    }

    return validStatus && extractStatus ? SUCCESS : FAILURE;
}

/**
 * Create new BDOC container.
 * Example command: <code>demo create --file=file1.txt --file=file2.txt --profile=TM demo-container.bdoc</code>
 * Args:
 * <ul>
 *   <li><b>--file</b> - option can occur multiple times. File(s) to be signed</li>
 *   <li><b>--city</b> - optional, city of production place</li>
 *   <li><b>--state</b> - optional, state of production place</li>
 *   <li><b>--postalCode</b> - optional, postalCode of production place</li>
 *   <li><b>--country</b> - optional, country of production place</li>
 *   <li><b>--role</b> - option can occur multiple times. Signer role(s)</li>
 *   <li><b>--pkcs11Driver</b> - optional, default is '/usr/lib/opensc-pkcs11.so'. Path of PKCS11 driver.</li>
 *   <li><b>--pin</b> - optional, default asks pin from prompt</li>
 *   <li><b>--profile</b> - optional, default TM. Signing profile, available values: BES, TM</li>
 * </ul>
 *
 * @param argc number of command line arguments.
 * @param argv command line arguments.
 * @return FAILURE (1) - failure, SUCCESS (0) - success
 */
int create(int argc, char* argv[])
{
    std::string path, pkcs11Driver, pin, city, state, postalCode, country;
    std::vector<std::string> files, roles;
    BDoc::Type profile = BDoc::TM;
    int returnCode = FAILURE;

    // Parse command line arguments.
    for(int i = 2; i < argc; i++)
    {
        std::string arg( decodeParameter( argv[i] ) );

        if(arg.find("--file=") == 0)
        {
            files.push_back(arg.substr(7));
        }
        else if(arg.find("--profile=") == 0)
        {
            arg = arg.substr(10);
            if(std::string("BES") == arg)     { profile = BDoc::BES; }
            else if(std::string("TM") == arg) { profile = BDoc::TM;  }
            else
            {
                printUsage(argv[0]);
                return FAILURE;
            }
        }
        else if(arg.find("--pkcs11Driver=") == 0)
            pkcs11Driver = arg.substr(15);
        else if(arg.find("--pin=") == 0)
            pin = arg.substr(6);
        else if(arg.find("--city=") == 0)
            city = arg.substr(7);
        else if(arg.find("--state=") == 0)
            state = arg.substr(8);
        else if(arg.find("--postalCode=") == 0)
            postalCode = arg.substr(13);
        else if(arg.find("--country=") == 0)
            country = arg.substr(10);
        else if(arg.find("--role=") == 0)
            roles.push_back(arg.substr(7));
        else
            path = arg;
    }

    // No bdoc container path given or no files given to sign
    if(path.empty() || files.empty())
    {
        printUsage(argv[0]);
        return FAILURE;
    }

    try
    {
        if(pkcs11Driver.empty())
            pkcs11Driver = Conf::getInstance()->getPKCS11DriverPath();

        // Create new BDOC container.
        BDoc bdoc;

        // Add all the documents to the container.
        for(std::vector<std::string>::const_iterator iter = files.begin(); iter != files.end(); iter++)
        {
            bdoc.addDocument(Document(*iter, "file"));
        }

        // Initialize signer implementation.
        DemoEstEIDConsolePinSigner signer(pkcs11Driver, pin);

        // Add signature production place.
        if(!city.empty() || !state.empty() || !postalCode.empty() || !country.empty())
        {
            SignatureProductionPlace spp(city, state, postalCode, country);
            signer.setSignatureProductionPlace(spp);
        }

        // Add signer role(s).
        if(!roles.empty())
        {
            SignerRole role;
            role.claimedRoles = roles;
            signer.setSignerRole(role);
        }

        // Sign the BDOC container.
        bdoc.sign(&signer, profile);

        // Save the BDOC container.
        bdoc.saveTo(path);

        returnCode = SUCCESS;
    }
    catch(const BDocException& e)
    {
        ERR("Caught BDocException: %s", parseException(e).c_str());
    }
    catch(const IOException& e)
    {
        ERR("Caught IOException: %s", parseException(e).c_str());
    }
    catch(const SignException& e)
    {
        ERR("Caught SignException: %s", parseException(e).c_str());
    }
    catch(const SignatureException& e)
    {
        ERR("Caught SignatureException: %s", parseException(e).c_str());
    }
    catch(const Exception& e)
    {
        ERR("Caught Exception: %s", parseException(e).c_str());
    }
    catch(...)
    {
        ERR("Caught unknown exception");
    }

    return returnCode;
}

/**
 * Sign BDOC container.
 * Example command: <code>demo sign --profile=TM demo-container.bdoc</code>
 * Args:
 * <ul>
 *   <li><b>--city</b> - optional, city of production place</li>
 *   <li><b>--state</b> - optional, state of production place</li>
 *   <li><b>--postalCode</b> - optional, postalCode of production place</li>
 *   <li><b>--country</b> - optional, country of production place</li>
 *   <li><b>--role</b> - option can occur multiple times. Signer role(s)</li>
 *   <li><b>--pkcs11Driver</b> - optional, default is '/usr/lib/opensc-pkcs11.so'. Path of PKCS11 driver.</li>
 *   <li><b>--pin</b> - optional, default asks pin from prompt</li>
 *   <li><b>--profile</b> - optional, default TM. Signing profile, available values: BES, TM</li>
 * </ul>
 *
 * @param argc number of command line arguments.
 * @param argv command line arguments.
 * @return FAILURE (1) - failure, SUCCESS (0) - success
 */
int sign(int argc, char* argv[])
{
    std::string path, pkcs11Driver, pin, city, state, postalCode, country;
    std::vector<std::string> roles;
    BDoc::Type profile = BDoc::TM;
    int returnCode = FAILURE;

    // Parse command line arguments.
    for(int i = 2; i < argc; i++)
    {
        std::string arg( decodeParameter( argv[i] ) );

        if(arg.find("--profile=") == 0)
        {
            arg = arg.substr(10);
            if(std::string("BES") == arg)     { profile = BDoc::BES; }
            else if(std::string("TM") == arg) { profile = BDoc::TM;  }
            else
            {
                printUsage(argv[0]);
                return FAILURE;
            }
        }
        else if(arg.find("--pkcs11Driver=") == 0)
            pkcs11Driver = arg.substr(15);
        else if(arg.find("--pin=") == 0)
            pin = arg.substr(6);
        else if(arg.find("--city=") == 0)
            city = arg.substr(7);
        else if(arg.find("--state=") == 0)
            state = arg.substr(8);
        else if(arg.find("--postalCode=") == 0)
            postalCode = arg.substr(13);
        else if(arg.find("--country=") == 0)
            country = arg.substr(10);
        else if(arg.find("--role=") == 0)
            roles.push_back(arg.substr(7));
        else
            path = arg;
    }

    // No bdoc container path given.
    if(path.empty())
    {
        printUsage(argv[0]);
        return FAILURE;
    }

    try
    {
        if(pkcs11Driver.empty())
            pkcs11Driver = Conf::getInstance()->getPKCS11DriverPath();

        // Create new BDOC container.
        BDoc bdoc(path);

        // Initialize signer implementation.
        DemoEstEIDConsolePinSigner signer(pkcs11Driver, pin);

        // Add signature production place.
        if(!city.empty() || !state.empty() || !postalCode.empty() || !country.empty())
        {
            SignatureProductionPlace spp(city, state, postalCode, country);
            signer.setSignatureProductionPlace(spp);
        }

        // Add signer role(s).
        if(!roles.empty())
        {
            SignerRole role;
            role.claimedRoles = roles;
            signer.setSignerRole(role);
        }

        // Sign the BDOC container.
        bdoc.sign(&signer, profile);

        // Save the BDOC container.
        bdoc.saveTo(path);

        returnCode = SUCCESS;
    }
    catch(const BDocException& e)
    {
        ERR("Caught BDocException: %s", parseException(e).c_str());
    }
    catch(const IOException& e)
    {
        ERR("Caught IOException: %s", parseException(e).c_str());
    }
    catch(const SignException& e)
    {
        ERR("Caught SignException: %s", parseException(e).c_str());
    }
    catch(const SignatureException& e)
    {
        ERR("Caught SignatureException: %s", parseException(e).c_str());
    }
    catch(const Exception& e)
    {
        ERR("Caught Exception: %s", parseException(e).c_str());
    }
    catch(...)
    {
        ERR("Caught unknown exception");
    }

    return returnCode;
}

/**
 * Initialize EstEID signed with driver implementation.
 *
 * @param driver full path to the PKCS #11 driver (e.g. /usr/lib/opensc-pkcs11.so)
 * @throws SignException exception is thrown if the provided PKCS #11 driver
 *         loading failed.
 */
DemoEstEIDConsolePinSigner::DemoEstEIDConsolePinSigner(const std::string &driver, const std::string &pin) throw(SignException)
 : EstEIDConsolePinSigner(driver)
 , m_pin(pin)
{
}

std::string DemoEstEIDConsolePinSigner::getPin(const PKCS11Cert &certificate) throw(SignException)
{
    return m_pin.empty() ? EstEIDConsolePinSigner::getPin(certificate) : m_pin;
}

/**
 * Just show available signing certificates. Uses base class EstEIDSigner implementation
 * to choose the correct signing certificate.
 *
 * NB! If you wan to choose the certificate yourself use PKCS11Signer as base class
 * and implement the <code>selectSigningCertificate()</code> method.
 */
PKCS11Signer::PKCS11Cert DemoEstEIDConsolePinSigner::selectSigningCertificate(
        const std::vector<PKCS11Signer::PKCS11Cert> &certificates) const throw(SignException)
{
    printf("Available certificates:\n");
    for(std::vector<PKCS11Signer::PKCS11Cert>::const_iterator iter = certificates.begin(); iter != certificates.end(); iter++)
        printPKCS11Cert(*iter);
    PKCS11Signer::PKCS11Cert cert = EstEIDSigner::selectSigningCertificate(certificates);
    printf("Selected:\n");
    printPKCS11Cert(cert);
    return cert;
}

/**
 * Prints certificate information.
 *
 * @param cert X.509 certificate.
 */
void DemoEstEIDConsolePinSigner::printPKCS11Cert(const PKCS11Signer::PKCS11Cert& cert) const
{
    printf("  -------------------------------------------------------\n");
    printf("  token label:        %s\n", cert.token.label.c_str());
    printf("  token manufacturer: %s\n", cert.token.manufacturer.c_str());
    printf("  token model:        %s\n", cert.token.model.c_str());
    printf("  token serial Nr:    %s\n", cert.token.serialNr.c_str());
    printf("  label: %s\n", cert.label.c_str());
}
