#include "log.h"

#include "Conf.h"
#include "util/File.h"
#include "util/File_p.h"

#include <fstream>
#include <iomanip>
#include <memory.h>
#include <stdio.h>

using namespace digidoc;

/**
 * Formats string, use same syntax as <code>printf()</code> function.
 * Example implementation from:
 * http://www.senzee5.com/2006/05/c-formatting-stdstring.html
 *
 * @param fmt format of the string. Uses same formating as <code>printf()</code> function.
 * @param ... parameters for the string format.
 * @return returns formatted string.
 */
std::string Log::format(const char* fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    std::string s = formatArgList(fmt, args);
    va_end(args);
    return s;
}

/**
 * Helper method for string formatting.
 *
 * @param fmt format of the string. Uses same formating as <code>printf()</code> function.
 * @param args parameters for the string format.
 * @return returns formatted string.
 * @see Util::String::format(const char* fmt, ...)
 */
std::string Log::formatArgList(const char* fmt, va_list args)
{
    if(!fmt)
    {
        return "";
    }

    int result = -1;
    int length = 256;
    char* buffer = 0;
    while(result == -1)
    {
        if (buffer) delete [] buffer;
        buffer = new char [length + 1];
        memset(buffer, 0, length + 1);
        result = vsnprintf(buffer, length, fmt, args);
        length *= 2;
    }

    std::string s(buffer);
    delete [] buffer;
    return s;
}

void Log::out(LogType type, const char *file, unsigned int line, const char *format, ...)
{
    Conf *conf = Conf::getInstance();
    if(!conf || conf->getLogLevel() < type)
        return;

    std::ostream *o = &std::cout;
    std::fstream f;
    if(!conf->getLogFile().empty())
    {
        f_string enc = util::File::encodeName(conf->getLogFile());
        f.open(enc.c_str(), std::fstream::out|std::fstream::app);
        o = &f;
    }
    switch(type)
    {
    case ErrorType: *o << "ERROR"; break;
    case WarnType: *o << "WARN"; break;
    case InfoType: *o << "INFO"; break;
    case DebugType: *o << "DEBUG"; break;
    }
    *o << " [" << file << ":" << line << "] - ";

    va_list args;
    va_start(args, format);
    *o << formatArgList(format, args).c_str() << "\n";
    va_end(args);
}

void Log::dbgPrintfMemImpl(const char *msg, const void *ptr, size_t size, const char *file, int line)
{
    Conf *conf = Conf::getInstance();
    if(!conf || conf->getLogLevel() < DebugType)
        return;

    std::ostream *o = &std::cout;
    std::fstream f;
    if(!conf->getLogFile().empty())
    {
        f_string enc = util::File::encodeName(conf->getLogFile());
        f.open(enc.c_str(), std::fstream::out|std::fstream::app);
        o = &f;
    }

    const unsigned char *data = (const unsigned char*)ptr;
    *o << "DEBUG [" << file << ":" << line << "] - " << msg << " { ";
    for(size_t i = 0; i < size; ++i)
        *o << '%' << std::setfill('0') << std::setw(2) << std::hex << std::uppercase << static_cast<int>(data[i]) << ' ';
    *o << "}:" << size << "\n";
}
